/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */

package com.smartgwt.mobile.client;

import com.google.gwt.core.client.EntryPoint;
import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.MetaElement;
import com.google.gwt.dom.client.NodeList;
import com.smartgwt.mobile.client.util.Page;
import com.smartgwt.mobile.client.util.SC;


/**
 * Internal SmartGWT Mobile entry point class where framework level initialization code executes
 * before the user's EntryPoint is run.
 */
public class SmartGwtMobileEntryPoint implements EntryPoint {
    private static boolean initialized = false;

    private static native void init() /*-{

        // In JSNI, we may be passed GWT Java Object references.
        // These typically can not be directly manipulated -- this check will test for such objects.
        $wnd.SmartGWTMobile = $wnd.SmartGWTMobile || {};
        $wnd.SmartGWTMobile.isNativeJavaObject = function (object) {
            // From observation "typeof" reports "function" for native Java objects in Firefox in development mode
            // and in OmniWeb, in development mode
            // In all other browsers, typeof reports as "object"
            var type = typeof object;
            if (type != "function" && type != "object") return false;
            if (@com.smartgwt.mobile.client.util.JSOHelper::isJSO(Ljava/lang/Object;)(object)) return false;
            return true;
        };

        // Define isc.DataSource.create to run the static DataSource.fromConfig() method.
        // This allows eval() to parse JavaScript sent from the server's data source loader.
        // E.g.:
        //     isc.DataSource.create({
        //         ID: "someName",
        //         inheritsFrom: isc.DataSource.create({
        //             ID: "someName_inheritsFrom",
        //             fields:[
        //                 {name:"whatever", type:"text"}
        //             ]
        //         })
        //     });
        // Note:  The use of window, rather than $wnd, is intentional because eval() uses window
        // as the top-level scope.
        window.isc = window.isc || {};
        window.isc.DataSource =  window.isc.DataSource || {};
        window.isc.DataSource.create = $entry(@com.smartgwt.mobile.client.data.DataSource::fromConfig(Lcom/google/gwt/core/client/JavaScriptObject;));
    }-*/;

    private boolean assertionsEnabled = false;
    public void onModuleLoad() {
        // Added boolean init check flag because GWT for some reason invokes this entry point class twice in hosted mode
        // even though it appears only once in the load hierarchy. Check with GWT team.
        if (!initialized) {
            initialized = true;

            assert (assertionsEnabled = true) == true; // Intentional side effect.
            if (!assertionsEnabled) {
                SC.logWarn("WARNING: Assertions are not enabled. It is recommended to develop with assertions " +
                        "enabled because both GWT and SmartGWT.mobile use them to help find bugs in application code. " +
                        "To enable assertions, recompile the application with the -ea option passed to the GWT compiler.");
            }

            init();

            // Look for a viewport <meta> tag. If one is not found, then log a warning and add the recommended viewport <meta>.
            final NodeList<Element> metaElems = Document.get().getElementsByTagName("meta");
            final int metaElems_length = metaElems.getLength();
            int i;
            for (i = 0; i < metaElems_length; ++i) {
                final MetaElement metaElem = metaElems.getItem(i).cast();
                if ("viewport".equals(metaElem.getName())) {
                    break;
                }
            }
            if (i >= metaElems_length) {
                SC.logWarn("No viewport <meta> tag was found. The viewport <meta> tag is essential to making a SmartGWT.mobile app look like " +
                        "a native app rather than a webpage packaged as an app. You can learn more about the viewport <meta> tag at " +
                        "https://developer.apple.com/library/safari/#documentation/AppleApplications/Reference/SafariHTMLRef/Articles/MetaTags.html. " +
                        "The recommended viewport <meta> tag <meta name=\"viewport\" content=\"initial-scale=1, width=device-width, user-scalable=no, minimum-scale=1, maximum-scale=1\"> will be added to the document.");
                Page._updateViewport(Float.valueOf(1.0f), "device-width", null, Boolean.FALSE);
            }
        }
    }
}