package com.smartgwt.mobile.client.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import com.smartgwt.mobile.client.types.OperatorId;

public class Criterion extends Criteria {

    private String fieldName;
    private OperatorId operator;
    private Object value;
    private List<Criterion> criteria;
    private Comparable<?> start, end;

    public Criterion() {
        criteria = new ArrayList<Criterion>();
    }

    public Criterion(Criterion c) {
        this.fieldName = c.fieldName;
        this.operator = c.operator;
        this.value = c.value;
        this.criteria = c.criteria;
        this.start = c.start;
        this.end = c.end;
    }

    public Criterion(OperatorId operator, Criterion[] criteria) {
        this(operator);
        this.criteria.addAll(Arrays.asList(criteria));
    }

    public Criterion(OperatorId operator) {
        this();
        this.operator = operator;
    }

    public Criterion(String fieldName, OperatorId operator) {
        this();
        this.fieldName = fieldName;
        this.operator = operator;
    }

    public Criterion(String fieldName, OperatorId operator, Object value) {
        this(fieldName, operator);
        this.value = value;
    }

    private <T> Criterion(String fieldName, OperatorId operator, Comparable<T> start, Comparable<T> end) {
        this(fieldName, operator);
        assert operator == OperatorId.BETWEEN || operator == OperatorId.BETWEEN_INCLUSIVE : "The operator type must be OperatorId.BETWEEN or OperatorId.BETWEEN_INCLUSIVE";
        this.start = start;
        this.end = end;
    }

    public Criterion(String fieldName, OperatorId operator, Byte start, Byte end) {
        this(fieldName, operator, (Comparable<Byte>)start, (Comparable<Byte>)end);
    }

    public Criterion(String fieldName, OperatorId operator, Short start, Short end) {
        this(fieldName, operator, (Comparable<Short>)start, (Comparable<Short>)end);
    }

    public Criterion(String fieldName, OperatorId operator, Integer start, Integer end) {
        this(fieldName, operator, (Comparable<Integer>)start, (Comparable<Integer>)end);
    }

    public Criterion(String fieldName, OperatorId operator, Long start, Long end) {
        this(fieldName, operator, (Comparable<Long>)start, (Comparable<Long>)end);
    }

    public Criterion(String fieldName, OperatorId operator, Float start, Float end) {
        this(fieldName, operator, (Comparable<Float>)start, (Comparable<Float>)end);
    }

    public Criterion(String fieldName, OperatorId operator, Double start, Double end) {
        this(fieldName, operator, (Comparable<Double>)start, (Comparable<Double>)end);
    }

    public Criterion(String fieldName, OperatorId operator, String start, String end) {
        this(fieldName, operator, (Comparable<String>)start, (Comparable<String>)end);
    }

    public Criterion(String fieldName, OperatorId operator, Date start, Date end) {
        this(fieldName, operator, (Comparable<Date>)start, (Comparable<Date>)end);
    }

    public final String getFieldName() {
        return fieldName;
    }

    public void setFieldName(String fieldName) {
        this.fieldName = fieldName;
    }

    public final OperatorId getOperator() {
        return operator;
    }

    public void setOperator(OperatorId operator) {
        this.operator = operator;
    }

    public final Object getValue() {
        return value;
    }

    public void setValue(Object value) {
        this.value = value;
    }

    public final Criterion[] getCriteria() {
        return criteria.toArray(new Criterion[criteria.size()]);
    }

    public final Comparable<?> getStart() {
        return start;
    }

    public final Comparable<?> getEnd() {
        return end;
    }

    private void appendToCriterionList(Criterion c) {
        c.unmarkAdvancedCriteria();
        criteria.add(c);
    }

    private void buildCriterionFromList(OperatorId operator, Criterion... criteria) {
        this.operator = operator;
        this.criteria = new ArrayList<Criterion>();
        for (Criterion c : criteria) {
            appendToCriterionList(c);
        }
    }

    public void addCriteria(Criterion c) {
        if (operator == OperatorId.AND) {
            appendToCriterionList(c);
        } else {
            final Criterion thisCopy = new Criterion(this);
            this.fieldName = null;
            this.value = null;
            this.start = this.end = null;
            buildCriterionFromList(OperatorId.AND, thisCopy, c);
        }
    }

    public void addCriteria(String field, Object value) {
        addCriteria(field, OperatorId.EQUALS, value);
    }

    public void addCriteria(String field, OperatorId operator, Object value) {
        addCriteria(new Criterion(field, operator, value));
    }

    public void markAdvancedCriteria() {
        setAttribute("_constructor", "AdvancedCriteria");
    }

    public void unmarkAdvancedCriteria() {
        remove("_constructor");
    }
}
