package com.smartgwt.mobile.client.internal.widgets;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.google.gwt.dom.client.Document;
import com.google.gwt.event.shared.HandlerRegistration;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.util.ElementUtil;
import com.smartgwt.mobile.client.internal.widgets.events.ContentChangedEvent;
import com.smartgwt.mobile.client.internal.widgets.events.ContentChangedHandler;
import com.smartgwt.mobile.client.internal.widgets.layout.NavigationBarItem;
import com.smartgwt.mobile.client.widgets.Canvas;
import com.smartgwt.mobile.client.widgets.NavigationButton;
import com.smartgwt.mobile.client.widgets.NavigationButton.NavigationDirection;

@SGWTInternal
public class NavigationButtons extends Canvas implements ContentChangedHandler, NavigationBarItem {

    private List<NavigationButton> buttons;
    private Set<HandlerRegistration> contentChangedRegistrations;

    public NavigationButtons(NavigationButton... buttons) {
        super(Document.get().createDivElement());
        setClassName("sc-navbuttons");

        this.buttons = new ArrayList<NavigationButton>();
        this.buttons.addAll(Arrays.asList(buttons));

        contentChangedRegistrations = new HashSet<HandlerRegistration>();
        for (NavigationButton button : buttons) {
            button.setDirection(NavigationDirection.NONE);
            addChild(button);
            contentChangedRegistrations.add(button._addContentChangedHandler(this));
        }
    }

    @Override
    public void destroy() {
        assert contentChangedRegistrations != null;
        for (HandlerRegistration contentChangedRegistration : contentChangedRegistrations) {
            contentChangedRegistration.removeHandler();
        }
        contentChangedRegistrations.clear();
        assert buttons != null;
        for (NavigationButton button : buttons) {
            removeChild(button);
        }
        buttons.clear();
        super.destroy();
    }

    @Override
    public final float _getContentWidth() {
        int ret = 0;
        for (NavigationButton button : buttons) {
            ret += ElementUtil.getOuterWidth(button.getElement(), true);
        }
        return ret;
    }

    public final NavigationButton getFirstButton() {
        if (buttons.isEmpty()) return null;
        return buttons.get(0);
    }

    @Override
    public void _onContentChanged(ContentChangedEvent event) {
        final Object source = event.getSource();
        if (buttons.contains(source)) {
            ContentChangedEvent.fire(this, this);
        }
    }

    @Override
    public HandlerRegistration _addContentChangedHandler(ContentChangedHandler handler) {
        return addHandler(handler, ContentChangedEvent.getType());
    }
}
