package com.smartgwt.mobile.client.internal.widgets;

import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.Style.Display;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.dom.client.UListElement;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.Window;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.EventUtil;
import com.smartgwt.mobile.client.internal.widgets.events.HasPickerHiddenHandlers;
import com.smartgwt.mobile.client.internal.widgets.events.HasValuesSelectedHandlers;
import com.smartgwt.mobile.client.internal.widgets.events.PickerHiddenEvent;
import com.smartgwt.mobile.client.internal.widgets.events.PickerHiddenHandler;
import com.smartgwt.mobile.client.internal.widgets.events.ValuesSelectedEvent;
import com.smartgwt.mobile.client.internal.widgets.events.ValuesSelectedHandler;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.widgets.Canvas;
import com.smartgwt.mobile.client.widgets.Panel;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.client.widgets.events.ClickHandler;
import com.smartgwt.mobile.client.widgets.layout.HLayout;
import com.smartgwt.mobile.client.widgets.layout.Layout;
import com.smartgwt.mobile.client.widgets.toolbar.ToolStrip;
import com.smartgwt.mobile.client.widgets.toolbar.ToolStripButton;
import com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants;

@SGWTInternal
public class Picker extends Panel implements HasValuesSelectedHandlers, HasPickerHiddenHandlers {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.pickerCSS().getText());
    }

    private static final int SLOT_HEIGHT = 44;

    private static class Column extends Layout {

        private LinkedHashMap<?, String> valueMap;
        private int offsetY, lastY = 0;
        private UListElement ul;

        protected Column(LinkedHashMap<?, String> valueMap) {
            this.valueMap = valueMap;
	        ul = Document.get().createULElement();
            if (valueMap != null && !valueMap.isEmpty()) {
                final StringBuilder lis = new StringBuilder();
                for (Map.Entry<?, String> value : valueMap.entrySet()) {
                    lis.append("<li>");
                    lis.append(value.getValue());
                    lis.append("</li>");
		        }
                ul.setInnerHTML(lis.toString());
	        }
            getElement().appendChild(ul);
		}

        public Map.Entry<?, String> getValueMapEntry(int pos) {
            if (valueMap == null || pos >= valueMap.size()) return null;
            final Iterator<?> it = valueMap.entrySet().iterator();
            while (true) {
                assert it.hasNext();
                @SuppressWarnings("unchecked")
                final Map.Entry<?, String> e = (Map.Entry<?, String>)it.next();
                if (pos-- == 0) return e;
            }
        }

        public int getPosition(Object value) {
            if (valueMap == null) return -1;
            int pos = 0;
            for (final Map.Entry<?, String> e : valueMap.entrySet()) {
                if (e.getKey() == null) {
                    if (value == null) return pos;
                } else if (e.getKey().equals(value)) return pos;
                ++pos;
            }
            return -1;
        }

        public void setPosition(int pos) {
            lastY = -pos * SLOT_HEIGHT;
            if (Canvas.isAndroid()) {
                getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translateY(" + lastY + "px)");
            } else {
                getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + lastY + "px, 0px)");
            }
        }

        private void setValue(Object value) {
            final int pos = Math.max(0, getPosition(value));
            setPosition(pos);
        }
	}
    private static class ReadOnlyColumn extends Column {
        public ReadOnlyColumn(LinkedHashMap<?, String> valueMap) {
            super(valueMap);
			this.setClassName("sc-right");
			this.setClassName("sc-readonly");
			this.setClassName("sc-shrink");
		}
	}
    private static class LeftColumn extends Column {
        public LeftColumn(LinkedHashMap<?, String> valueMap) {
            super(valueMap);
			this.setClassName("sc-left");
		}
	}
    private static class CenterColumn extends Column {
        public CenterColumn(LinkedHashMap<?, String> valueMap) {
            super(valueMap);
			this.setClassName("sc-center");
			this.setClassName("sc-layout-flex1");
		}
	}
    private static class RightColumn extends Column {
        public RightColumn(LinkedHashMap<?, String> valueMap) {
            super(valueMap);
			this.setClassName("sc-right");
		}
    }

    private long lastMouseWheelEventTime;
    private boolean visible = false, mouseWheelActive = false;;
    private boolean active = false, pickermoved = false;
    private Object[] selectedValues;
    private LinkedHashMap<?, String>[] valueMaps;
    private ToolStrip toolbar;
    private ToolStripButton cancel, done;
    private Canvas activeSlot;
    private Element selectionSlot;
    private HLayout frame, slotsWrapper, slots;
    private Column[] columns;
    private Panel screenCover;
    private int clientY = 0, activeSlotIndex = 0;
    private Layout parent;
    private boolean isHiding = false;
    private HandlerRegistration screenCoverClickRegistration, cancelClickRegistration,
            doneClickRegistration;

    public Picker(Layout parent, Layout screenCoverParent, LinkedHashMap... valueMaps) {
        if (valueMaps == null) throw new NullPointerException("valueMaps");
        if (valueMaps.length < 1 || 3 < valueMaps.length) throw new IllegalArgumentException("valueMaps");
        this.valueMaps = valueMaps;
        selectedValues = new Object[valueMaps.length];
        screenCover = new Panel();
        screenCover.setClassName("sc-screen-cover");
        screenCover.getElement().getStyle().setDisplay(Display.NONE);
        screenCoverClickRegistration = screenCover.addClickHandler(new ClickHandler() {

            @Override
            public void onClick(ClickEvent event) {
                if(Picker.this.isVisible()) {
                    Picker.this.hide();
                }
            }
            
        });
		this.parent = parent;
    	this.setClassName("sc-picker-wrapper");
        this.toolbar = new ToolStrip();
        this.toolbar.setTintColor("#333");
        this.toolbar.setClassName("sc-layout-box-pack-justify");
        cancel = new ToolStripButton("Cancel");
        cancel.removeClassName("default");
        cancel.removeClassName("action");
        cancel.setClassName("bordered");
        cancel.setClassName("sc-custom-tint");
        cancel.getElement().getStyle().setBackgroundColor("#333");
        cancelClickRegistration = cancel.addClickHandler(new ClickHandler() {
            @Override
            public void onClick(ClickEvent event) {
                if (!isHiding) hide();
            }
        });
        toolbar.addButton(cancel);
        done = new ToolStripButton("Done");
        done.removeClassName("default");
        done.removeClassName("action");
        done.setClassName("bordered");
        done.setClassName("important");
        doneClickRegistration = done.addClickHandler(new ClickHandler() {
            @Override
            public void onClick(ClickEvent event) {
                if (!isHiding) {
                    ValuesSelectedEvent.fire(Picker.this, selectedValues);
                    hide();
                }
            }
        });
        toolbar.addButton(done);
        this.addMember(toolbar);
        slotsWrapper = new HLayout();
        slotsWrapper.removeClassName("sc-layout-hbox");
        slotsWrapper.setClassName("sc-slots-wrapper");
        slots = new HLayout();
        slots.setClassName("sc-slots");
        slotsWrapper.addMember(slots);
        this.addMember(slotsWrapper);
        frame = new HLayout();
        frame.removeClassName("sc-layout-hbox");
        frame.setClassName("sc-picker-frame");
        Element child = Document.get().createDivElement();
        frame.getElement().appendChild(child);
        selectionSlot = Document.get().createDivElement();
        child.appendChild(selectionSlot);
        this.addMember(frame);
		int top = Window.getClientHeight() + Window.getScrollTop();
        getElement().getStyle().setTop(top, Unit.PX);
        final int size = valueMaps.length;
        columns = new Column[size];
        {
            switch (size) {
                case 1:
                    columns[0] = new CenterColumn(valueMaps[0]);
                    slots.addMember(columns[0]);
                    selectedValues[0] = valueMaps[0].isEmpty() ? null : columns[0].getValueMapEntry(0).getKey();
                    break;
                case 2:
                    columns[0] = new LeftColumn(valueMaps[0]);
                    slots.addMember(columns[0]);
                    columns[1] = new RightColumn(valueMaps[1]);
                    slots.addMember(columns[1]);
                    selectedValues[0] = valueMaps[0].isEmpty() ? null : columns[0].getValueMapEntry(0).getKey();
                    selectedValues[1] = valueMaps[1].isEmpty() ? null : columns[1].getValueMapEntry(0).getKey();
                    break;
                case 3:
                    columns[0] = new ReadOnlyColumn(valueMaps[0]);
                    slots.addMember(columns[0]);
                    columns[1] = new LeftColumn(valueMaps[1]);
                    slots.addMember(columns[1]);
                    columns[2] = new RightColumn(valueMaps[2]);
                    slots.addMember(columns[2]);
                    selectedValues[0] = valueMaps[0].isEmpty() ? null : columns[0].getValueMapEntry(0).getKey();
                    selectedValues[1] = valueMaps[1].isEmpty() ? null : columns[1].getValueMapEntry(0).getKey();
                    selectedValues[2] = valueMaps[2].isEmpty() ? null : columns[2].getValueMapEntry(0).getKey();
                    break;
            }
        }
    	getElement().getStyle().setDisplay(Display.NONE);
    	parent.addMember(this);
		sinkEvents(Event.ONCLICK | Event.GESTUREEVENTS | Event.TOUCHEVENTS | Event.MOUSEEVENTS | Event.FOCUSEVENTS | Event.ONMOUSEWHEEL | Event.ONSCROLL);		
	}

    @Override
    public void destroy() {
        if (columns != null) {
            for (Column column : columns) {
                column.destroy();
            }
            columns = null;
        }
        if (doneClickRegistration != null) {
            doneClickRegistration.removeHandler();
            doneClickRegistration = null;
        }
        if (cancelClickRegistration != null) {
            cancelClickRegistration.removeHandler();
            cancelClickRegistration = null;
        }
        if (screenCoverClickRegistration != null) {
            screenCoverClickRegistration.removeHandler();
            screenCoverClickRegistration = null;
        }
        screenCover.destroy();
        super.destroy();
    }

    public final Object[] getSelectedValues() {
        return selectedValues;
	}

    public void setSelectedValues(Object[] selectedValues) {
        if (selectedValues == null) selectedValues = new Object[columns.length];
        else if (selectedValues.length != columns.length) throw new IllegalArgumentException("selectedValues");
        assert selectedValues.length == columns.length;
        for (int i = 0; i < columns.length; ++i) {
            columns[i].setValue(selectedValues[i]);
        }
        this.selectedValues = selectedValues;
    }

    public final LinkedHashMap<?, String>[] getValueMaps() {
        return valueMaps;
	}

	public boolean isVisible() {
		return visible;
	}

    public void hide() {
        if (isHiding) return;
        isHiding = true;
        getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, 0px, 0px)");
		visible = false;
        new Timer() {
            public void run() {
                isHiding = false;
                screenCover.removeClassName("visible");
                screenCover.setClassName("hidden");
                getElement().getStyle().setDisplay(Display.NONE);
                parent.removeMember(screenCover);
                getElement().getStyle().setDisplay(Display.NONE);
                PickerHiddenEvent.fire(Picker.this, Picker.this);
            }
        }.schedule(320);
	}

    @Override
    public void onBrowserEvent(Event event) {
        super.onBrowserEvent(event);

        Element targetElem = EventUtil.getTargetElem(event);
        if (targetElem != null && frame.getElement().isOrHasChild(targetElem)) {
            switch(event.getTypeInt()) {
            case Event.ONMOUSEDOWN:
                if(!touched) {
                    onStart(event);
                }
                break;
            case Event.ONTOUCHSTART:
            case Event.ONGESTURESTART:
                touched = true;
                onStart(event);
                break;
            case Event.ONMOUSEWHEEL:
                if(!mouseWheelActive) {
                    mouseWheelActive = true;
                    onStart(event);
                } else {
                    onMouseWheelMove(event);
                }
                break;
            case Event.ONMOUSEMOVE:
                if(!touched){
                    onMove(event);
                }
                break;
            case Event.ONTOUCHMOVE:
            case Event.ONGESTURECHANGE:
            case Event.ONSCROLL:
                onMove(event);
                break;
            case Event.ONMOUSEUP:
                if(!touched) {
                    onEnd(event);
                }
                break;
            case Event.ONTOUCHEND:
            case Event.ONTOUCHCANCEL:
            case Event.ONGESTUREEND:
                onEnd(event);
                break;
            }
        }
    }

    private void setActiveSlot(int xPos) {
    	List<Canvas> slots = Arrays.asList(this.slots.getMembers());
    	int i = 0;
    	for(Canvas slot: slots) {
    		int right = slot.getElement().getAbsoluteRight();
    		int left = slot.getElement().getAbsoluteLeft();
    		if(xPos > left && xPos < right) {
    			activeSlot = slot;
    			activeSlotIndex = i;
    			break;
    		}
    		++i;
    	}
    }

	public void show() {
        parent.addMember(screenCover,parent.getMemberNumber(this));
        screenCover.getElement().getStyle().clearDisplay();
        screenCover.removeClassName("hidden");
        screenCover.setClassName("visible");
        getElement().getStyle().setDisplay(Display.BLOCK);
    	final int height = isIPhone() && !isStandAlone() ? 244 - 65 : 244;
        Timer timer = new Timer() {
            @Override
            public void run() {
                Picker.this.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, -" + height + "px, 0px)");
            }
        };
        visible = true;   
        timer.schedule(1);

	}

	public void onStart(Event event) {
    	event.preventDefault();
    	event.stopPropagation();
    	if(!active) {
        	active = true;
        	pickermoved = false;
        	setActiveSlot(event.getTypeInt() == Event.ONTOUCHSTART ? event.getTouches().get(0).getClientX() : event.getClientX());
        	clientY = (event.getTypeInt() == Event.ONTOUCHSTART ? event.getTouches().get(0).getClientY() : event.getClientY());
        	((Column)activeSlot).offsetY = clientY - ((Column)activeSlot).lastY;
    	}		
	}

	public void onMouseWheelMove(final Event event) {
		if (!isEnabled()) {
			return;
		}
		if(active) {
			event.preventDefault();
			event.stopPropagation();
			if(activeSlot != null) {
				pickermoved = true;
				int deltaY = getWheelDelta(event);
				((Column)activeSlot).lastY += deltaY;
        		int height = ((Column)activeSlot).ul.getOffsetHeight() - activeSlot.getOffsetHeight();
        		if(((Column)activeSlot).lastY < 0 && Math.abs(((Column)activeSlot).lastY) > height) {
        			((Column)activeSlot).lastY = -height;
        		} else if(((Column)activeSlot).lastY > 0) {
        			((Column)activeSlot).lastY = 0;
        		}
        		if(Canvas.isAndroid()) {
                    activeSlot.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translateY(" + ((Column)activeSlot).lastY + "px)");
        		} else {
                    activeSlot.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + ((Column)activeSlot).lastY + "px, 0px)");
        		}
				lastMouseWheelEventTime = new Date().getTime();
				new Timer() {
			        public void run() {
						if(Picker.this.mouseWheelActive) {
							long currentTime = new Date().getTime();
							final int mouseWheelDeltaTime = new Long(currentTime - lastMouseWheelEventTime).intValue();
							if(mouseWheelDeltaTime > 80) {
								Picker.this.onEnd(event);
								Picker.this.mouseWheelActive = false;
							}
						}			        
					}
			    }.schedule(80);
			} 
		}
	}

	public void onMove(Event event) {
    	event.preventDefault();
    	event.stopPropagation();
    	if(active) {
        	if(activeSlot != null) {
        		pickermoved = true;
        		((Column)activeSlot).lastY = (event.getTypeInt() == Event.ONTOUCHMOVE ? event.getTouches().get(0).getClientY() : event.getClientY()) - ((Column)activeSlot).offsetY;
        		int height = ((Column)activeSlot).ul.getOffsetHeight() - activeSlot.getOffsetHeight();
        		if(((Column)activeSlot).lastY < 0 && Math.abs(((Column)activeSlot).lastY) > height) {
        			((Column)activeSlot).lastY = -height;
        		} else if(((Column)activeSlot).lastY > 0) {
        			((Column)activeSlot).lastY = 0;
        		}
                if(Canvas.isAndroid()) {
                    activeSlot.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translateY(" + ((Column)activeSlot).lastY + "px)");
                } else {
                    activeSlot.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + ((Column)activeSlot).lastY + "px, 0px)");
                }
            } 	
    	}	
	}

    public void onEnd(Event event) {
        if (active) {
            active = false;
            if (!pickermoved) {
                int selectionSlotY = selectionSlot.getAbsoluteBottom() - 22;
                int deltaY = clientY - selectionSlotY;
                if(deltaY > 0) {
                    ((Column)activeSlot).lastY -= SLOT_HEIGHT;
                } else {
                    ((Column)activeSlot).lastY += SLOT_HEIGHT;
                }
                activeSlot.getElement().addClassName("slide");
                new Timer() {
                    @Override
                    public void run() {
                        if (activeSlot != null && activeSlot.getElement() != null) {
                            activeSlot.getElement().removeClassName("slide");
                        }
                    }
                }.schedule(320);
            } 
            float position = Math.abs(((Column)activeSlot).lastY) / ((float)SLOT_HEIGHT);
            int pos = Math.round(position);
            ((Column)activeSlot).setPosition(pos);
            final Map.Entry<?, String> e = ((Column)activeSlot).getValueMapEntry(pos);
            selectedValues[activeSlotIndex] = e == null ? null : e.getKey();
            activeSlot = null;
            activeSlotIndex = 0;
            clientY = 0; 
            pickermoved = false;
        }
    }

    @Override
    public HandlerRegistration _addValuesSelectedHandler(ValuesSelectedHandler handler) {
        return addHandler(handler, ValuesSelectedEvent.getType());
    }

    @Override
    public HandlerRegistration _addPickerHiddenHandler(PickerHiddenHandler handler) {
        return addHandler(handler, PickerHiddenEvent.getType());
    }
}
