package com.smartgwt.mobile.client.internal.widgets;

import java.util.ArrayList;
import java.util.List;

import com.google.gwt.dom.client.DivElement;
import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.dom.client.Style.Display;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.ui.RootLayoutPanel;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.i18n.SmartGwtMessages;
import com.smartgwt.mobile.client.internal.types.PromptType;
import com.smartgwt.mobile.client.internal.util.UsernameAndPasswordCallback;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.types.DefaultDialogAction;
import com.smartgwt.mobile.client.util.BooleanCallback;
import com.smartgwt.mobile.client.util.IsCallback;
import com.smartgwt.mobile.client.util.ValueCallback;
import com.smartgwt.mobile.client.widgets.BaseButton;
import com.smartgwt.mobile.client.widgets.Button;
import com.smartgwt.mobile.client.widgets.Dialog;
import com.smartgwt.mobile.client.widgets.Panel;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.client.widgets.events.ClickHandler;
import com.smartgwt.mobile.client.widgets.form.DynamicForm;
import com.smartgwt.mobile.client.widgets.form.fields.PasswordItem;
import com.smartgwt.mobile.client.widgets.form.fields.TextItem;
import com.smartgwt.mobile.client.widgets.toolbar.ToolStripButton;

@SGWTInternal
public final class PromptDialog extends Dialog {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.alertviewCSS().getText());
    }

    public static final PromptDialog INSTANCE = new PromptDialog();

    private interface OpDetails {}

    private static class ShowingDetails implements OpDetails {

        public final PromptType type;
        public final String title;
        public final String message;
        public final IsCallback userCallback;

        public ShowingDetails(PromptType type, String title, String message, IsCallback userCallback) {
            this.type = type;
            this.title = title;
            this.message = message;
            this.userCallback = userCallback;
        }

        public void onDialogClose() {
            if (userCallback != null) {
                switch (type) {
                    case NOTICE:
                    case YES_OR_NO:
                        ((BooleanCallback)userCallback).execute(null);
                        break;
                    case CONFIRM:
                        ((BooleanCallback)userCallback).execute(null);
                        break;
                    case PROMPT:
                        break;
                    case PROMPT_PLAIN_TEXT:
                    case PROMPT_SECURE_TEXT:
                        ((ValueCallback)userCallback).execute(null);
                        break;
                    case PROMPT_USERNAME_AND_PASSWORD:
                        ((UsernameAndPasswordCallback)userCallback).execute(null, null);
                        break;
                }
            }
        }

        @Override
        public String toString() {
            String message = this.message == null ? null : (this.message.length() > 15 ? (this.message.substring(0, 12) + "...") : this.message);
            return "{ " + type.toString() + ", title = " + (title == null ? "null" : "'" + title + "'") +
                    ", message = " + (message == null ? "null" : "'" + message + "'") +
                    ", userCallback = " + userCallback +
                    " }";
        }
    }

    private static class HidingDetails implements OpDetails {

        public static final HidingDetails INSTANCE = new HidingDetails();

        private HidingDetails() {}
    }

    //private DivElement outerWrapperDiv; // Use `this.getElement().<DivElement>cast()`
    private DivElement wrapperDiv;
    private DivElement alertviewDiv;

    private DynamicForm form;
    private TextItem textItem;
    private PasswordItem passwordItem;

    private DivElement alertviewButtonsDiv;
    private BaseButton firstButton;
    private BaseButton secondButton;

    private Button okButton;

    // A queue of operations that are either in-progress or pending.
    private final List<OpDetails> opDetailsQueue = new ArrayList<OpDetails>();
    // The ShowingDetails corresponding to the PromptDialog on screen and fully-shown (not
    // in the process of being shown).
    private ShowingDetails currentShowingDetails = null;

    private PromptDialog() {
        final Document document = Document.get();

        final RootLayoutPanel rootLayoutPanel = RootLayoutPanel.get();
        final Panel screenCover = _getScreenCover();
        screenCover.getElement().addClassName("sc-screen-alertview-cover");
        rootLayoutPanel.add(screenCover);
        // Adding a Widget to GWT's root layout panel sets the `display` property to "absolute"
        // The next line sets the `display` of `screenCover` back to "none".
        screenCover.getElement().getStyle().setDisplay(Display.NONE);

        final DivElement outerWrapperDiv = DivElement.as(getElement());
        outerWrapperDiv.removeClassName("sc-action-sheet");
        outerWrapperDiv.setClassName("sc-alertview-outer-wrapper");
        outerWrapperDiv.getStyle().setDisplay(Display.NONE);
        wrapperDiv = document.createDivElement();
        wrapperDiv.setClassName("sc-alertview-wrapper");
        outerWrapperDiv.appendChild(wrapperDiv);
        alertviewDiv = document.createDivElement();
        alertviewDiv.setClassName("sc-alertview");
        wrapperDiv.appendChild(alertviewDiv);
        removeMember(header);
        add(header, alertviewDiv.<com.google.gwt.user.client.Element>cast());
        final Panel contentPanel = _getContentPanel();
        removeMember(contentPanel);
        add(contentPanel, alertviewDiv.<com.google.gwt.user.client.Element>cast());

        form = new DynamicForm();
        textItem = new TextItem("text");
        textItem.setShowTitle(false);
        passwordItem = new PasswordItem("password");
        passwordItem.setShowTitle(false);
        add(form, alertviewDiv.<com.google.gwt.user.client.Element>cast());

        alertviewButtonsDiv = document.createDivElement();
        alertviewButtonsDiv.setClassName("sc-alertview-buttons");
        alertviewDiv.appendChild(alertviewButtonsDiv);
        final ClickHandler firstAndSecondButtonClickHandler = new ClickHandler() {

            @Override
            public void onClick(ClickEvent event) {
                if (! _isHiding() && currentShowingDetails != null) {
                    final ShowingDetails details = currentShowingDetails;
                    currentShowingDetails = null;
                    assert details.type != PromptType.NOTICE;
                    assert details.type != PromptType.PROMPT;
                    hide();
                    assert event.getSource() == firstButton || event.getSource() == secondButton;
                    switch (details.type) {
                        case YES_OR_NO:
                            if (details.userCallback != null) {
                                ((BooleanCallback)details.userCallback).execute(event.getSource() == firstButton ? Boolean.TRUE : Boolean.FALSE);
                            }
                            break;
                        case CONFIRM:
                            if (details.userCallback != null) {
                                ((BooleanCallback)details.userCallback).execute(event.getSource() == firstButton ? Boolean.TRUE : null);
                            }
                            break;
                        case PROMPT_PLAIN_TEXT:
                        case PROMPT_SECURE_TEXT:
                            if (details.userCallback != null) {
                                ValueCallback userCallback = (ValueCallback)details.userCallback;
                                if (event.getSource() == firstButton) {
                                    final TextItem textItem = details.type == PromptType.PROMPT_PLAIN_TEXT ? PromptDialog.this.textItem : passwordItem;
                                    final String value = textItem.getValueAsString();
                                    userCallback.execute(value == null ? "" : value);
                                } else {
                                    userCallback.execute(null);
                                }
                            }
                            break;
                        case PROMPT_USERNAME_AND_PASSWORD:
                            if (details.userCallback != null) {
                                UsernameAndPasswordCallback userCallback = (UsernameAndPasswordCallback)details.userCallback;
                                if (event.getSource() == firstButton) {
                                    final String username = textItem.getValueAsString();
                                    final String password = passwordItem.getValueAsString();
                                    userCallback.execute(username == null ? "" : username, password == null ? "" : password);
                                } else {
                                    userCallback.execute(null, null);
                                }
                            }
                            break;

                        case NOTICE:
                        case PROMPT:
                            assert false;
                            break;
                    }
                }
            }
        };
        firstButton = new ToolStripButton();
        firstButton.addClickHandler(firstAndSecondButtonClickHandler);
        firstButton._getLabelElement().addClassName("h1");
        secondButton = new ToolStripButton();
        secondButton.addClickHandler(firstAndSecondButtonClickHandler);
        secondButton._getLabelElement().addClassName("h1");
        add(secondButton, alertviewButtonsDiv.<com.google.gwt.user.client.Element>cast());
        add(firstButton, alertviewButtonsDiv.<com.google.gwt.user.client.Element>cast());

        okButton = new Button(SmartGwtMessages.INSTANCE.dialog_OkButtonTitle());
        okButton.addClickHandler(new ClickHandler() {

            @Override
            public void onClick(ClickEvent event) {
                if (! _isHiding() && currentShowingDetails != null) {
                    final ShowingDetails details = currentShowingDetails;
                    currentShowingDetails = null;
                    assert details.type == PromptType.NOTICE;
                    hide();
                    if (details.userCallback != null) {
                        ((BooleanCallback)details.userCallback).execute(Boolean.TRUE);
                    }
                }
            }
        });
        add(okButton, alertviewDiv.<com.google.gwt.user.client.Element>cast());

        rootLayoutPanel.add(this);
        getElement().getStyle().setDisplay(Display.NONE);
    }

    @SuppressWarnings("unused")
    private String opDetailsQueueToString() {
        assert opDetailsQueue != null;
        final StringBuilder sb = new StringBuilder();
        sb.append("[ ");
        final int opDetailsQueue_size = opDetailsQueue.size();
        for (int i = 0; i < opDetailsQueue_size; ++i) {
            if (i != 0) sb.append(", ");
            sb.append(opDetailsQueue.get(i));
        }
        if (opDetailsQueue_size == 0) sb.append(']');
        else sb.append(" ]");
        return sb.toString();
    }

    @Override
    protected void _onAnimationEnd(NativeEvent event, String animationName, double elapsedTime) {
        super._onAnimationEnd(event, animationName, elapsedTime);
        if ("sc-alertview_fadeIn".equals(animationName)) {
            _onShown();
        } else if ("sc-alertview_fadeOut".equals(animationName)) {
            _onHidden();
        }
    }

    @Override
    public void destroy() {
        assert false : "PromptDialog.destroy() should never be called.";
    }

    @Override
    protected void onLoad() {
        final RootLayoutPanel rootLayoutPanel = RootLayoutPanel.get();
        final Element screenCoverContainerElement = rootLayoutPanel.getWidgetContainerElement(_getScreenCover());
        assert screenCoverContainerElement != null;
        final Element containerElement = rootLayoutPanel.getWidgetContainerElement(this);
        assert containerElement != null;

        new Timer() {
            @Override
            public void run() {
                screenCoverContainerElement.getStyle().setDisplay(Display.NONE);
                containerElement.getStyle().setDisplay(Display.NONE);
            }
        }.schedule(1);
    }

    @Override
    public void hide() {
        assert (_isShowing() || _isHiding()) == ! opDetailsQueue.isEmpty();
        opDetailsQueue.add(HidingDetails.INSTANCE);
        if (opDetailsQueue.size() != 1) {
            return;
        }
        super.hide();
    }

    public void show(final PromptType type, final String title, final String message, IsCallback userCallback) {
        assert (_isShowing() || _isHiding()) == ! opDetailsQueue.isEmpty();
        final ShowingDetails details = new ShowingDetails(type, title, message, userCallback);
        if (opDetailsQueue.isEmpty()) {
            if (isVisible()) {
                assert currentShowingDetails != null;
                currentShowingDetails.onDialogClose();
                currentShowingDetails = null;

                // Hide the dialog that is on screen to show the new details.
                opDetailsQueue.add(HidingDetails.INSTANCE);
                opDetailsQueue.add(details);
                hide();
            } else {
                opDetailsQueue.add(details);
                callShow(details);
            }
        } else opDetailsQueue.add(details);
    }

    private void callShow(ShowingDetails details) {
        assert ! isVisible();
        String title = details.title;
        setTitle(title == null || (title = title.trim()).isEmpty() ? "&nbsp;" : title);
        String message = details.message;
        setMessage(message == null ? "" : message);
        final PromptType type = details.type;
        switch (type) {
            case NOTICE:
                form.setVisible(false);
                alertviewButtonsDiv.getStyle().setDisplay(Display.NONE);
                okButton.setVisible(true);
                break;
            case YES_OR_NO:
                form.setVisible(false);
                alertviewButtonsDiv.getStyle().clearDisplay();
                firstButton.setTitle(SmartGwtMessages.INSTANCE.dialog_YesButtonTitle());
                secondButton.setTitle(SmartGwtMessages.INSTANCE.dialog_NoButtonTitle());
                okButton.setVisible(false);
                break;
            case CONFIRM:
                form.setVisible(false);
                alertviewButtonsDiv.getStyle().clearDisplay();
                firstButton.setTitle(SmartGwtMessages.INSTANCE.dialog_OkButtonTitle());
                secondButton.setTitle(SmartGwtMessages.INSTANCE.dialog_CancelButtonTitle());
                okButton.setVisible(false);
                break;
            case PROMPT:
                form.setVisible(false);
                alertviewButtonsDiv.getStyle().setDisplay(Display.NONE);
                okButton.setVisible(false);
                break;
            case PROMPT_PLAIN_TEXT:
            case PROMPT_SECURE_TEXT:
            case PROMPT_USERNAME_AND_PASSWORD:
                form.setVisible(true);
                alertviewButtonsDiv.getStyle().clearDisplay();
                if (type == PromptType.PROMPT_PLAIN_TEXT) {
                    form.setItems(textItem);
                } else if (type == PromptType.PROMPT_SECURE_TEXT) {
                    form.setItems(passwordItem);
                } else {
                    assert type == PromptType.PROMPT_USERNAME_AND_PASSWORD;
                    form.setItems(textItem, passwordItem);
                }
                firstButton.setTitle(SmartGwtMessages.INSTANCE.dialog_OkButtonTitle());
                secondButton.setTitle(SmartGwtMessages.INSTANCE.dialog_CancelButtonTitle());
                okButton.setVisible(false);
                break;
        }
        super.show();
    }

    private void reset() {
        setTitle(null);
        setMessage(null);
        textItem.setValue(null);
        passwordItem.setValue(null);
        firstButton.setTitle(null);
        secondButton.setTitle(null);
    }

    @Override
    protected void _doShow() {
        //SC.logWarn("PromptDialog._doShow()");
        final RootLayoutPanel rootLayoutPanel = RootLayoutPanel.get();
        final Element screenCoverContainerElement = rootLayoutPanel.getWidgetContainerElement(_getScreenCover());
        assert screenCoverContainerElement != null;
        screenCoverContainerElement.getStyle().clearDisplay();
        final Element containerElement = rootLayoutPanel.getWidgetContainerElement(this);
        assert containerElement != null;
        containerElement.getStyle().clearDisplay();
    }

    @Override
    protected void _onShown() {
        //SC.logWarn("PromptDialog._onShown(): opDetailsQueue = " + opDetailsQueueToString());
        super._onShown();
        assert ! opDetailsQueue.isEmpty();
        assert opDetailsQueue.get(0) instanceof ShowingDetails;
        currentShowingDetails = (ShowingDetails)opDetailsQueue.remove(0);
        if (!opDetailsQueue.isEmpty()) {
            currentShowingDetails.onDialogClose();
            currentShowingDetails = null;

            final OpDetails nextOp = opDetailsQueue.get(0);
            if (nextOp instanceof HidingDetails) {
                super.hide();
            } else {
                assert nextOp instanceof ShowingDetails;
                opDetailsQueue.add(0, HidingDetails.INSTANCE);
                super.hide();
            }
        }
    }

    @Override
    protected void _doHide() {}

    @Override
    protected void _onHidden() {
        //SC.logWarn("PromptDialog._onHidden(): opDetailsQueue = " + opDetailsQueueToString());
        super._onHidden();
        assert ! isVisible();
        final RootLayoutPanel rootLayoutPanel = RootLayoutPanel.get();
        final Element screenCoverContainerElement = rootLayoutPanel.getWidgetContainerElement(_getScreenCover());
        assert screenCoverContainerElement != null;
        screenCoverContainerElement.getStyle().setDisplay(Display.NONE);
        final Element containerElement = rootLayoutPanel.getWidgetContainerElement(this);
        assert containerElement != null;
        containerElement.getStyle().setDisplay(Display.NONE);
        reset();

        assert opDetailsQueue.get(0) instanceof HidingDetails;
        do {
            opDetailsQueue.remove(0);
        } while (!opDetailsQueue.isEmpty() && opDetailsQueue.get(0) instanceof HidingDetails);
        if (!opDetailsQueue.isEmpty()) {
            final OpDetails nextOp = opDetailsQueue.get(0);
            assert nextOp instanceof ShowingDetails;
            callShow((ShowingDetails)nextOp);
        }
    }

    @Override
    public void setDefaultAction(DefaultDialogAction defaultAction) {
        assert defaultAction == DefaultDialogAction.HIDE;
    }
}
