package com.smartgwt.mobile.client.internal.widgets.form;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.LIElement;
import com.google.gwt.dom.client.LabelElement;
import com.google.gwt.dom.client.TableCellElement;
import com.google.gwt.dom.client.TableElement;
import com.google.gwt.dom.client.TableRowElement;
import com.google.gwt.dom.client.TableSectionElement;
import com.google.gwt.dom.client.UListElement;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.theme.iphone.FormCssResourceIPhone;
import com.smartgwt.mobile.client.internal.theme.iphone.ThemeResourcesIPhone;
import com.smartgwt.mobile.client.util.SC;
import com.smartgwt.mobile.client.widgets.form.DynamicForm;
import com.smartgwt.mobile.client.widgets.form.fields.FormItem;
import com.smartgwt.mobile.client.widgets.form.fields.HiddenItem;
import com.smartgwt.mobile.client.widgets.form.fields.SearchItem;
import com.smartgwt.mobile.client.widgets.form.fields.SliderItem;
import com.smartgwt.mobile.client.widgets.form.fields.SpacerItem;
import com.smartgwt.mobile.client.widgets.form.fields.SwitchItem;
import com.smartgwt.mobile.client.widgets.form.fields.TextAreaItem;

@SGWTInternal
public class DynamicFormImplIPhone extends DynamicFormImpl {

    private static final FormCssResourceIPhone CSS = ThemeResourcesIPhone._INSTANCE.formCSS();
    static {
        CSS.ensureInjected();
    }

    private final Map<String, Element> errors = new HashMap<String, Element>();
    private TableElement table = null;

    @Override
    public void destroyImpl(DynamicForm self) {
        // TODO Auto-generated method stub

    }

    @Override
    public void setFields(DynamicForm self, FormItem[] items) {
        if (table != null && table.getParentNode() != null) table.removeFromParent();
        final Map<String, FormItem> fields = self._getFieldsMap();
        assert fields.isEmpty();
        if (items != null) {
            Document document = Document.get();
            table = document.createTableElement();
            table.setCellPadding(0);
            table.setCellSpacing(0);
            table.setClassName(CSS.formClass());
            TableSectionElement tbody = document.createTBodyElement();
            table.appendChild(tbody);
            TableRowElement firstTR = null, lastTR = null;
            for (int i = 0; i < items.length; ++i) {
                final FormItem item = items[i];
                item._setForm(self);
                final String name = item.getName();
                fields.put(name, item);

                if (item instanceof HiddenItem) continue;

                TableRowElement tr = document.createTRElement();
                if (firstTR == null) {
                    tr.addClassName(CSS.firstClass());
                    firstTR = tr;
                }
                tbody.appendChild(tr);

                TableCellElement th;
                TableCellElement td = document.createTDElement();
                td.setClassName(CSS.formItemCellClass());
                if (!(item instanceof SpacerItem) && item.getShowTitle()) {
                    th = document.createTHElement();
                    th.setClassName(CSS.labelCellClass());
                    th.addClassName(CSS.firstClass());
                    tr.appendChild(th);
                } else {
                    th = null;
                    td.addClassName(CSS.firstClass());
                }
                td.addClassName(CSS.lastClass());
                if (item instanceof SpacerItem) {
                    if (lastTR != null) lastTR.addClassName(CSS.followedBySpacerItemRowClass());
                    tr.addClassName(CSS.spacerItemRowClass());
                } else if (item instanceof SearchItem) {
                    if (th != null) th.addClassName(CSS.searchItemLabelCellClass());
                    td.addClassName(CSS.searchItemCellClass());
                } else if (item instanceof SliderItem) {
                    if (th != null) th.addClassName(CSS.sliderItemLabelCellClass());
                } else if (item instanceof SwitchItem) {
                    if (th != null) th.addClassName(CSS.switchItemLabelCellClass());
                } else if (item instanceof TextAreaItem) {
                    if (item.getShowTitle()) {
                        if (th != null) {
                            th.addClassName(CSS.textAreaItemLabelCellClass());
                            th.setColSpan(2);
                            th.addClassName(CSS.lastClass());
                        }
                        tr = document.createTRElement();
                        tbody.appendChild(tr);
                    }
                    td.addClassName(CSS.textAreaItemCellClass());
                    td.setColSpan(2);
                    td.addClassName(CSS.firstClass());
                }
                tr.appendChild(td);
                self._add(item, td.<com.google.gwt.user.client.Element>cast());
                if (!(item instanceof SpacerItem) && item.getShowTitle()) {
                    assert th != null;
                    LabelElement label = document.createLabelElement();
                    label.setHtmlFor(item.getElement().getId());
                    label.setInnerHTML(item.getTitle());
                    th.appendChild(label);
                } else td.setColSpan(2);

                lastTR = tr;
            }
            if (lastTR != null) lastTR.addClassName(CSS.lastClass());
            self.getElement().appendChild(table);
        }
    }

    @Override
    public void clearErrors(DynamicForm self) {
        final Map<String, FormItem> fields = self._getFieldsMap();
        for (Map.Entry<String, Element> e : errors.entrySet()) {
            final Element error = e.getValue();
            error.removeFromParent();
            fields.get(e.getKey()).getElement().removeClassName(CSS.formItemValueInvalidClass());
        }
        errors.clear();
    }

    @Override
    public void showError(DynamicForm self, boolean showError, FormItem field, Object errorObj) {
        Element error = null;
        if (errorObj instanceof String) {
            final String errorMessage = (String)errorObj;
            error = Document.get().createDivElement();
            error.setInnerHTML(errorMessage);
        } else if (errorObj instanceof List || errorObj instanceof String[]) {
            final int numMessages = errorObj instanceof List ? ((List<?>)errorObj).size() : ((String[])errorObj).length;
            if (numMessages <= 1) {
                if (numMessages == 1) {
                    final String errorMessage;
                    if (errorObj instanceof List) {
                        final Object obj = ((List<?>)errorObj).get(0);
                        if (obj == null) errorMessage = "&nbsp;";
                        else errorMessage = obj.toString();
                    } else errorMessage = ((String[])errorObj)[0];
                    error = Document.get().createDivElement();
                    error.setInnerHTML(errorMessage);
                }
            } else {
                final UListElement ul = Document.get().createULElement();
                error = ul;
                for (int i = 0; i < numMessages; ++i) {
                    final LIElement li = Document.get().createLIElement();
                    final String errorMessage;
                    if (errorObj instanceof List) {
                        final Object obj = ((List<?>)errorObj).get(i);
                        if (obj == null) errorMessage = "&nbsp;";
                        else errorMessage = obj.toString();
                    } else errorMessage = ((String[])errorObj)[i];
                    li.setInnerHTML(errorMessage);
                    ul.appendChild(li);
                }
            }
        } else {
            SC.logWarn("Unhandled error object type " + errorObj.getClass().getName() + " for field '" + field.getName() + "'");
        }

        if (error != null && showError) {
            error.setClassName(CSS.errorLabelClass());
            errors.put(field.getName(), error);
            if (showError) {
                final Element itemElement = field.getElement();
                itemElement.getParentElement().insertAfter(error, itemElement);
            }
        }
    }
}
