/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */

package com.smartgwt.mobile.client.widgets;

import java.util.Iterator;

import com.google.gwt.core.client.GWT;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.dom.client.Style;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.event.dom.client.TouchMoveEvent;
import com.google.gwt.event.dom.client.TouchMoveHandler;
import com.google.gwt.event.logical.shared.ResizeEvent;
import com.google.gwt.event.logical.shared.ResizeHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.ComplexPanel;
import com.google.gwt.user.client.ui.RootLayoutPanel;
import com.google.gwt.user.client.ui.Widget;
import com.google.gwt.user.client.ui.WidgetCollection;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.EventHandler;
import com.smartgwt.mobile.client.internal.EventUtil;
import com.smartgwt.mobile.client.internal.util.ElementUtil;
import com.smartgwt.mobile.client.internal.widgets.CanvasStaticImpl;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.util.SC;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.client.widgets.events.ClickHandler;
import com.smartgwt.mobile.client.widgets.events.HasClickHandlers;
import com.smartgwt.mobile.client.widgets.events.HasShowContextMenuHandlers;
import com.smartgwt.mobile.client.widgets.events.ShowContextMenuEvent;
import com.smartgwt.mobile.client.widgets.events.ShowContextMenuHandler;
import com.smartgwt.mobile.client.widgets.grid.ListGrid;
import com.smartgwt.mobile.client.widgets.layout.Layout;
import com.smartgwt.mobile.client.widgets.menu.Menu;

/**
 * Canvas is the base class for SmartGWT.mobile widgets.
 */
public class Canvas extends ComplexPanel implements HasClickHandlers, HasShowContextMenuHandlers {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.baseCSS().getText());

        EventHandler.maybeInit();
        SC._maybeInit();

        RootLayoutPanel.get().addDomHandler(new TouchMoveHandler(){
            @Override
            public void onTouchMove(TouchMoveEvent event) {
                if (Canvas.useIOSNativeScrolling()) {
                    final Element targetElem = EventUtil.getElement(event.getNativeEvent().getEventTarget());
                    for (Element el = targetElem; el != null; el = el.getParentElement()) {
                        if (ElementUtil.hasClassName(el, "sc-scrollable")) return;
                    }
                }

                event.preventDefault();
            }
        }, TouchMoveEvent.getType());
        Window.addResizeHandler(new ResizeHandler() {
            @Override
            public void onResize(ResizeEvent event) {
                Canvas._maybeHideAddressBar();
            }
        });
    }

    private static final CanvasStaticImpl IMPL = GWT.create(CanvasStaticImpl.class);

    @SGWTInternal
    public static final boolean _HISTORY_ENABLED = IMPL.isHistoryEnabled();

    @SGWTInternal
    protected static boolean touched = false;

    protected String id;
    protected String className;
    private boolean disabled = false;
    private Object width; // Integer or String
    private Object height; // Integer or String
    private Menu contextMenu = null;

	protected native int parseDimension(String dim) /*-{
        if (dim == "auto" || dim == "inherit" || dim == "") {
          return 0;
        } else {
          dim = dim.replace(/^(\s*[+-]?((\d+\.?\d*)|(\.\d+))([eE][+-]?\d+)?)(.*)$/, "$1");
          return parseInt(dim);
        }
    }-*/;
	protected native String split(String path, String token, int part) /*-{
        var parts = path.split(token);
        if(parts && parts.length && (parts.length > part)) {
            return unescape(parts[part]);
        }
        return "";
    }-*/;
    public static native boolean isStandAlone() /*-{
        return $wnd.navigator.standalone || false;
    }-*/;
    @SGWTInternal
    protected static void _maybeHideAddressBar() {
        if (isIPhone() && !isStandAlone() && !isUIWebView()) {
            new Timer() {
                public void run() {
                    final Style style = RootLayoutPanel.get().getElement().getStyle();
                    style.setHeight(Window.getClientHeight() + 100 - 40, Style.Unit.PX);
                    Window.scrollTo(0,0);
                }
            }.schedule(100);
        } else {
            final Style style = RootLayoutPanel.get().getElement().getStyle();
            style.setHeight(Window.getClientHeight(), Style.Unit.PX);
        }
    }

    public native int getComputedY(Element element) /*-{
        var computedStyle = $wnd.getComputedStyle(element, null);
        var transformValue = computedStyle[@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::INSTANCE.@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::getTransformPropertyName()()];
        if (transformValue != null) {
            var parts = $wnd.String(transformValue).replace(/[^0-9-.,]/g, '').split(',');
            if (parts.length >= 6) {
                var ret = $wnd.parseInt(parts[5]);
                if (!$wnd.isNaN(ret)) {
                    return ret;
                }
            }
        }
        return 0;
    }-*/;

    public static boolean isAndroid() {
        return IMPL.isAndroid();
    }

    public static boolean isIPad() {
        return IMPL.isIPad();
    }
    public static boolean isIPhone() {
        return IMPL.isIPhone();
    }
    public static boolean isUIWebView() {
        return IMPL.isUIWebView();
    }
    public static boolean useIOSNativeScrolling() {
        return IMPL.useIOSNativeScrolling();
    }

    public Canvas() {}

    public Canvas(Element element) {
        this();
        setElement(element);
    }

    @Override
    public void onBrowserEvent(Event event) {
        switch (event.getTypeInt()) {
            case Event.ONCLICK:
                if (disabled) return;

                final boolean cancelled = ClickEvent._fire(this, event.getClientX(), event.getClientY());
                if (cancelled) event.stopPropagation();
                break;

            case Event.ONCONTEXTMENU:
                if (disabled) return;

                // If this is a ListGrid, then exit early because ListGrid implementations
                // have more specialized context menu support that incorporates information
                // about which row and possibly which cell was clicked.
                if (this instanceof ListGrid) return;

                if (contextMenu != null) {
                    final boolean showContextMenuCancelled = ShowContextMenuEvent._fire(this);
                    if (!showContextMenuCancelled) {
                        final int x = event.getClientX(),
                                y = event.getClientY();
                        contextMenu._showAt(this, x, y, x, x, y, y);
                    }
                }
                break;
        }
    }

    @SGWTInternal
    public com.google.gwt.user.client.Element _getInnerElement() {
        return getElement();
    }

    public void addChild(Canvas canvas) {
        removeChild(canvas);
        add(canvas, _getInnerElement());
    }
    
    public void addChild(int index, Canvas canvas) {
        removeChild(canvas);
        insert(canvas, getElement(), index, true);
    }

    public void addChild(Widget widget) {
        removeChild(widget);
        add(new WidgetCanvas(widget), _getInnerElement());
    }

    public final boolean hasChild(Canvas canvas) {
        return getChildren().contains(canvas);
    }

    public final boolean hasChild(Widget widget) {
        WidgetCollection children = getChildren();
        Iterator<Widget> it = children.iterator();
        while (it.hasNext()) {
            Widget child = it.next();
            if (child instanceof WidgetCanvas) {
                if (((WidgetCanvas) child)._getWidget().equals(widget)) {
                    return true;
                }
            }
        }
        return false;
    }

    public void removeChild(Canvas canvas) {
        remove(canvas);
    }

    public void removeChild(Widget widget) {
        WidgetCollection children = getChildren();
        Iterator<Widget> it = children.iterator();
        while (it.hasNext()) {
            Widget child = it.next();
            if (child instanceof WidgetCanvas) {
                if (((WidgetCanvas) child)._getWidget().equals(widget)) {
                    it.remove();
                    return;
                }
            }
        }
    }

    public String getClassName() {
		return className;
	}

	public void setClassName(String className) {
		this.setClassName(className, false);
	}

	public void setClassName(String className, boolean replace) {
		this.className = className;
		if(replace) {
			getElement().setClassName(className);
		} else {
			getElement().addClassName(className);
		}
	}

	public void removeClassName(String className) {
		this.className = null;
		getElement().removeClassName(className);
	}

	/**
     * Return the ID of the Canvas.
     *
     * @return the ID
     */
    public String getId() {
        return id;
    }

    /**
     * Set the Canvas ID
     *
     * @param id the ID
     */
    public void setId(String id) {
        this.id = id;
        this.getElement().setId(this.id);
    }

    public String getInnerHTML() {
        return _getInnerElement().getInnerHTML();
    }

    public void setContents(String html) {
        _getInnerElement().setInnerHTML(html);
    }
    

    /**
     * Is the button enabled?
     *
     * @return true if enabled
     */
    public boolean isEnabled() {
        return disabled == false;
    }

    /**
     * Set the enabled state of this canvas.
     *
     * @param onoff true to enable, false to disable. Default is true
     */
    public void setDisabled(boolean onoff) {
        this.disabled = onoff;
        if (onoff) {
            disable();
        } else {
            enable();
        }
    }

    public void destroy() {
		try {
			Widget parent = getParent();
			if(parent instanceof Layout) {
				((Layout)parent).removeMember(this);
			}
            removeFromParent(); // Automatically calls onDetach().
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

    /**
     * Enable the button. Shortcut to setEnabled(true).
     */
    public void enable() {
        getElement().removeClassName("disabled");
        getElement().setPropertyBoolean("disabled", false);
        this.disabled = false;
    }

    /**
     * Disable the button. Shortcut to setEnabled(false).
     */
    public void disable() {
        getElement().addClassName("disabled");
        getElement().setPropertyBoolean("disabled", true);
        this.disabled = true;
    }

    public void setMargin(int margin) {
    	getElement().getStyle().setMargin(margin, Style.Unit.PX);
    }

    public int getMargin() {
    	return new Integer(getElement().getStyle().getMargin());
    }

    public final Integer getWidth() {
        return (Integer) width;
    }

    public void setWidth(int width) {
        this.width = Integer.valueOf(width);
    }

    public final Integer getHeight() {
        return (Integer) height;
    }

    public void setHeight(int height) {
        this.height = Integer.valueOf(height);
    }

    public final Menu getContextMenu() {
        return contextMenu;
    }

    public void setContextMenu(Menu contextMenu) {
        if (this.contextMenu != null) this.contextMenu._hide();
        this.contextMenu = contextMenu;
        sinkEvents(Event.ONCONTEXTMENU);
    }

    @Override
    public HandlerRegistration addClickHandler(ClickHandler handler) {
        sinkEvents(Event.ONCLICK);
        return addHandler(handler, ClickEvent.getType());
    }

    @Override
    public HandlerRegistration addShowContextMenuHandler(ShowContextMenuHandler handler) {
        return addHandler(handler, ShowContextMenuEvent.getType());
    }
}
