package com.smartgwt.mobile.client.widgets;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.ImageElement;
import com.google.gwt.dom.client.Style;
import com.google.gwt.dom.client.TableCellElement;
import com.google.gwt.dom.client.TableElement;
import com.google.gwt.dom.client.TableRowElement;
import com.google.gwt.dom.client.TableSectionElement;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.user.client.ui.Image;
import com.smartgwt.mobile.client.types.ImageStyle;
import com.smartgwt.mobile.client.util.SC;

public class Img extends Canvas {

    private TableSectionElement tbody;
    private TableRowElement tr;
    private TableCellElement td;
    private Image image;

    private ImageStyle imageType;
    private Integer imageWidth;
    private Integer imageHeight;

    private Img(Image image) {
        createElements();
        this.image = image;
        image.getElement().<ImageElement>cast().setAttribute("align", "top");
        add(image, td.<com.google.gwt.user.client.Element>cast());
        setWidth(100);
        setHeight(100);
        setImageType(ImageStyle.STRETCH);
    }

    public Img() {
        this(new Image());
    }

    public Img(String src) {
        this(new Image(src));
    }

    public Img(String src, int width, int height) {
        this(src);
        setWidth(width);
        setHeight(height);
    }

    public Img(ImageResource resource) {
        this(new Image(resource));
    }

    private void createElements() {
        final Document document = Document.get();
        final TableElement table = document.createTableElement();
        table.addClassName("sc-img");
        table.setBorder(0);
        table.setCellPadding(0);
        table.setCellSpacing(0);
        tbody = document.createTBodyElement();
        table.appendChild(tbody);
        tr = document.createTRElement();
        tbody.appendChild(tr);
        td = document.createTDElement();
        tr.appendChild(td);
        setElement(table);
    }

    @Override
    public void setWidth(int width) {
        super.setWidth(width);
        getElement().getStyle().setWidth(width, Style.Unit.PX);
        td.getStyle().setWidth(width, Style.Unit.PX);
    }

    @Override
    public void setHeight(int height) {
        super.setHeight(height);
        td.getStyle().setHeight(height, Style.Unit.PX);
    }

    public final String getAltText() {
        return image.getElement().<ImageElement>cast().getAlt();
    }

    public void setAltText(String altText) {
        image.getElement().<ImageElement>cast().setAlt(altText);
    }

    public final Integer getImageWidth() {
        return imageWidth;
    }

    public void setImageWidth(Integer imageWidth) {
        this.imageWidth = imageWidth;
        if (imageType != null && imageType != ImageStyle.NORMAL && imageType != ImageStyle.CENTER) {
            SC.warn("setImageWidth() called on an Img widget with imageType:" + imageType.getValue());
            return;
        }
        if (imageWidth == null) {
            image.setWidth("auto");
        } else {
            image.setWidth(imageWidth.intValue() + "px");
        }
    }

    public final Integer getImageHeight() {
        return imageHeight;
    }

    public void setImageHeight(Integer imageHeight) {
        this.imageHeight = imageHeight;
        if (imageType != null && imageType != ImageStyle.NORMAL && imageType != ImageStyle.CENTER) {
            SC.warn("setImageHeight() called on an Img widget with imageType:" + imageType.getValue());
            return;
        }
        if (imageHeight == null) {
            image.setHeight("auto");
        } else {
            image.setHeight(imageHeight.intValue() + "px");
        }
    }

    public final ImageStyle getImageType() {
        return imageType;
    }

    public void setImageType(ImageStyle imageType) {
        this.imageType = imageType;
        if (imageType == null) {
            imageType = ImageStyle.NORMAL;
        }
        switch (imageType) {
            case NORMAL:
                td.setAlign("left");
                td.setVAlign("top");
                if (imageWidth == null) {
                    image.setWidth("auto");
                } else {
                    image.setWidth(imageWidth.intValue() + "px");
                }
                if (imageHeight == null) {
                    image.setHeight("auto");
                } else {
                    image.setHeight(imageHeight.intValue() + "px");
                }
                break;
            case CENTER:
                td.setAlign("center");
                td.setVAlign("middle");
                if (imageWidth == null) {
                    image.setWidth("auto");
                } else {
                    image.setWidth(imageWidth.intValue() + "px");
                }
                if (imageHeight == null) {
                    image.setHeight("auto");
                } else {
                    image.setHeight(imageHeight.intValue() + "px");
                }
                break;
            case STRETCH:
                td.setAlign("left");
                td.setVAlign("top");
                image.setWidth("100%");
                image.setHeight("100%");
                break;
        }
    }
}
