package com.smartgwt.mobile.client.widgets;

import com.smartgwt.mobile.client.internal.widgets.NavigationButtons;
import com.smartgwt.mobile.client.internal.widgets.events.ContentChangedEvent;
import com.smartgwt.mobile.client.internal.widgets.layout.NavigationBarItem;
import com.smartgwt.mobile.client.widgets.toolbar.ToolStripButton;

public class NavigationButton extends ToolStripButton implements NavigationBarItem {

    private NavigationDirection direction = NavigationDirection.NONE;

    public NavigationButton(String title) {
        super(title);
    }

    public NavigationButton(String title, NavigationDirection direction) {
        this(title);
        setDirection(direction);
    }

    public NavigationButton(NavigationDirection direction) {
        super();
        setDirection(direction);
    }

    public NavigationDirection getDirection() {
        return direction;
    }

    public void setDirection(NavigationDirection direction) {
        if (direction == null) direction = NavigationDirection.NONE;
        if (getParent() instanceof NavigationButtons && direction != NavigationDirection.NONE) {
            throw new IllegalArgumentException("Only NavigationDirection NONE is allowed when multiple buttons are in use.");
        }
        if (this.direction != direction) {
            getElement().removeClassName(this.direction.getValue());
            this.direction = direction;

            if (this.direction != NavigationDirection.NONE) {
                getElement().addClassName(this.direction.getValue());
                if (tintColor != null) {
                    setTintColor(tintColor);
                }
            }
            ContentChangedEvent.fire(this, this);
        }
    }

    /**
     * Should this button render with a specific navigation direction or as a square action button?
     */
    public enum NavigationDirection {
        /**
         * No navigation direction - renders as a square button
         */
        NONE("none"),

        /**
         * Navigate forward - renders as an arrow pointing right
         */
         FORWARD("next"),

        /**
         * Navigate backwards - renders as an arrow pointing left
         */
        BACK("back");

        private String value;

        NavigationDirection(String value) {
            this.value = value;
        }

        public String getValue() {
            return value;
        }
    }
}
