package com.smartgwt.mobile.client.widgets;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.History;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.types.Alignment;
import com.smartgwt.mobile.client.types.PanelHeaderPosition;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.client.widgets.events.ClickHandler;
import com.smartgwt.mobile.client.widgets.layout.Layout;
import com.smartgwt.mobile.client.widgets.layout.NavStack;
import com.smartgwt.mobile.client.widgets.toolbar.ToolStrip;
import com.smartgwt.mobile.client.widgets.toolbar.ToolStripButton;

public class Panel extends Layout {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.panelCSS().getText());
    }

	protected List<Action> actions = new ArrayList<Action>();
    private PanelContainer container;
	protected PanelHeader header;
	protected boolean showPanelHeader = false;
    public native int getWheelDelta(Event event) /*-{
		return event.wheelDelta;
	}-*/;
    protected void create(String title, NavStack<?> nav) {
    }
    public void parseHistory(final NavStack<?> nav) {
        String historyToken = History.getToken();
        String path = split(historyToken,"/",nav.size());
        if(path != null && path.length() > 0) {
            create(path, nav);
        }
    }

	public Panel() {
        this.header = new PanelHeader(null);
	}

	public Panel(String title) {
		this.header = new PanelHeader(title);
	}

	public Panel(String title, ImageResource icon) {
		this.header = new PanelHeader(title, icon);
	}

    public Action[] getActions() {
        return actions.toArray(new Action[]{});
	}

    public void setActions(Action actions[]) {
        this.actions = Arrays.asList(actions);
	}

    public final PanelContainer getPanelContainer() {
        return container;
    }

	public PanelHeaderPosition getHeaderPosition() {
		return header.getHeaderPosition();
	}

	public void setHeaderPosition(PanelHeaderPosition headerPosition) {
		this.header.setHeaderPosition(headerPosition);
	}

	public String getIconStyle() {
		return this.header.getIconStyle();
	}

	public void setIconStyle(String iconStyle) {
		this.header.setIconStyle(iconStyle);
	}

	public boolean isShowPanelHeader() {
		return showPanelHeader;
	}

	public void setShowPanelHeader(boolean showPanelHeader) {
		this.showPanelHeader = showPanelHeader;
	}

	public String getTitle() {
		return this.header.getTitle();
	}

	public void setTitle(String title) {
		this.header.setTitle(title);
	}
	
	public ImageResource getIcon() {
		return this.header.getIcon();
	}

	public void setIcon(ImageResource icon) {
		this.header.setIcon(icon);
	}

	public void containerChanged(PanelContainer container) {
		this.container = container;
		ContainerFeatures features = container.getContainerFeatures();
		if(showPanelHeader || (actions != null && actions.size() > features.getMaxIconControls())) {
	        ToolStrip toolbar = new ToolStrip();
	        toolbar.setAlign(Alignment.RIGHT);
	        if(!features.isShowsTitle()) {
	            Header1 title = new Header1(header.getTitle());
	            toolbar.addMember(title);
	        }
	        if(!features.isShowsPrimaryAction() && actions != null && actions.size() > 0) {
	            toolbar.addSpacer();
	            for(final Action action: getActions()) {
	                final ToolStripButton button = new ToolStripButton(action.getTitle());
	                button.addClickHandler(new ClickHandler() {
	                    @Override
	                    public void onClick(ClickEvent event) {
	                        action.execute(new ActionContext() {
	                            @Override
	                            public Panel getPanel() {
	                                return Panel.this;
	                            }
	                            @Override
	                            public Canvas getControl() {
	                                return button;
	                            }                           
	                        });
	                    }            
	                });
	                toolbar.addButton(button);
	            }
	        }
	        addMember(toolbar,0);
		}
	}
}
