package com.smartgwt.mobile.client.widgets;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.Style.Display;
import com.google.gwt.dom.client.StyleInjector;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.client.widgets.events.ClickHandler;
import com.smartgwt.mobile.client.widgets.layout.HLayout;
import com.smartgwt.mobile.client.widgets.layout.Layout;
import com.smartgwt.mobile.client.widgets.layout.VLayout;
import com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants;

public class Popover extends Dialog {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.popoverCSS().getText());
    }
    Canvas target;

	private native void popover(Element target, Element menu) /*-{
    	var targetLeft = target.offsetLeft;
    	var targetBottom = target.offsetHeight;
    	var parent = target.parentNode;
    	var roffset = 0, aoffset = 0;
    	while(parent && !isNaN(parent.offsetTop)) {
    		if(parent.style.position === 'absolute' && !aoffset) {
    			aoffset = parent.offsetTop;
    		} else if(parent.style.position === 'relative') {
    			roffset += parent.offsetTop;
    		}
    		parent = parent.parentNode;
    	}
    	targetBottom += roffset + aoffset;
    	var targetWidth = target.offsetWidth;
    	var menuWidth = menu.offsetWidth || 320;
    	var menuLeft = targetLeft + (targetWidth/2) - (menuWidth/2);
    	menuLeft = Math.max(menuLeft,0);
    	menu.style.display = 'block';
    	menu.style.top = (targetBottom + 15) + 'px';
    	var screenWidth = window.innerWidth || (window.screen && window.screen.width);
    	var rightLimit = menuLeft + menuWidth;
    	if(rightLimit > screenWidth) {
    	    menu.style.left = (screenWidth - 5) + "px";
    	} else {
    	    menu.style.left = menuLeft + 'px';
    	}
	}-*/;
	Layout parent;
	public Popover(Layout parent, String title, Canvas content) {
		setClassName("sc-popover", true);
        getElement().getStyle().setDisplay(Display.NONE);
        Element pointer = Document.get().createDivElement();
        pointer.addClassName("sc-popover-pointer-top");
        pointer.addClassName("left");
        getElement().appendChild(pointer);
		HLayout header = new HLayout();
		header.setClassName("sc-header");
		header.removeClassName("sc-layout-hbox");
	    Button cancel = new Button("Cancel");
		cancel.setClassName("bordered");
		cancel.setClassName("sc-custom-tint");
		cancel.removeClassName("default");
		cancel.removeClassName("action");
        cancel.addClickHandler(new ClickHandler() {
            public void onClick(ClickEvent event) {
                Popover.this.hide();
            }
        });
		Header1 center = new Header1(title);
		Button done = new Button("Done");
		done.setClassName("bordered");
		done.setClassName("important");
		done.removeClassName("default");
		done.removeClassName("action");
        done.addClickHandler(new ClickHandler() {
            public void onClick(ClickEvent event) {
                Popover.this.hide();
            }
        });
		header.addMember(cancel);
		header.addMember(center);
		header.addMember(done);
        addMember(header);
		VLayout section = new VLayout();
		section.setClassName("sc-section");
		section.addMember(content);
        addMember(section);
        parent.addMember(this);
        if(screenCover != null) {
            screenCover.addClickHandler(new ClickHandler() {

                @Override
                public void onClick(ClickEvent event) {
                    Popover.this.hide();
                }
                
            });

        }
        this.parent = parent;
	}

	@Override
	public void hide() {
        screenCover.removeClassName("visible");
        screenCover.setClassName("hidden");
		getElement().getStyle().setDisplay(Display.NONE);
        parent.removeMember(screenCover);

	}
	
	public void show(Canvas target) {
        parent.addMember(screenCover,parent.getMemberNumber(this));
        screenCover.getElement().getStyle().clearDisplay();
        screenCover.removeClassName("hidden");
        screenCover.setClassName("visible");
		popover(target.getElement(), this.getElement());
        getElement().getStyle().setProperty(DOMConstants.INSTANCE.getAnimationShorthandPropertyName(), "fadeInAnim 0.35s 1");
	}

}
