package com.smartgwt.mobile.client.widgets;

import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants;

class ScrollablePanelImplIOSNative extends ScrollablePanelImpl {

    static {
        // http://johanbrook.com/browsers/native-momentum-scrolling-ios-5/
        StyleInjector.injectAtStart(".sc-scrollable{position:absolute;width:100%;height:100%;overflow-x:hidden;overflow-y:auto}" +

                // http://cantina.co/2012/03/06/ios-5-native-scrolling-grins-and-gothcas/
                ".sc-scrollable>.sc-inner-scrollable{-webkit-transform:translate3d(0,0,0)}" +

                // There is a weird bug with iOS native scrolling where the use of transforms
                // and -webkit-overflow-scrolling:touch can disable all events outside of the
                // content area of the scrollable element. See: http://stackoverflow.com/questions/10862771/webkit-transform-translate3d-webkit-overflow-scrolling-touch-no-click-events
                // (Note: Although this Stack Overflow question only mentions the bug when using 3D
                // transforms, the bug still occurs when using 2D transforms.)
                // Also: http://stackoverflow.com/questions/7763458/ios5-webkit-overflow-scrolling-causes-touch-events-stopping-work
                //
                // One work-around is to position the scrollable element so that it takes up
                // the entire screen. For example, with a 47px-tall NavigationBar at the top,
                // the ScrollablePanel element would be styled position:absolute;top:0;right:0;bottom:0;left:0;padding-top:47px
                // and the NavigationBar would be absolutely positioned on top.
                // The problem with this work-around is that the scrollbar is oddly cut off
                // when scrolling to the top and it will likely be very difficult to use this
                // work-around in general because all other elements need to be absolutely
                // positioned on top of the scrollable element. Another potential problem is
                // that possibly only one element can be scrollable at any given time.
                //
                // Another work-around seems to be: (1) Style the <html> element overflow:auto.
                // (2) Before running an animation that uses transforms, set the -webkit-overflow-scrolling
                // CSS property of the <html> element to the default value of 'auto' (3) After
                // the animation finishes, set the -webkit-overflow-scrolling CSS property of
                // the <html> element to 'touch'.
                // Through experimentation in the iOS Simulator, it appears to be necessary to
                // set the -webkit-overflow-scrolling CSS property of the <html> element in
                // timeouts. See the code in AnimationUtil.java.
                // In iOS Simulator, this trick appears to fix the "no events" issue, but occasionally
                // and seemingly randomly, there is a new problem that the browser refuses to scroll
                // the overflow content and all attempts to scroll the scrollable content result
                // in the entire screen shifting in the direction of the scroll attempt.
                // TODO Fix this issue.
                "html{overflow:auto}"
                );
    }

    private boolean preventNativeScrolling = false;

    @Override
    void init(ScrollablePanel self) {
        self.sinkEvents(Event.ONTOUCHSTART | Event.ONTOUCHMOVE);
    }

    @Override
    void onLoad(final ScrollablePanel self) {
        super.onLoad(self);
        new Timer() {
            @Override
            public void run() {
                self.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getOverflowScrollingPropertyName(), "touch");
            }
        }.schedule(1);
    }

    @Override
    void onUnload(ScrollablePanel self) {
        super.onUnload(self);
    }

    // Use Joe Lambert's ScrollFix technique, slightly modified
    // http://blog.joelambert.co.uk/2011/10/14/a-fix-for-ios5-overflow-scrolling-scrollfix-js/
    @Override
    void onBrowserEvent(ScrollablePanel self, Event event) {
        self.superOnBrowserEvent(event);
        final Element elem = self.getElement();
        switch (event.getTypeInt()) {
            case Event.ONTOUCHSTART:
                final int scrollTop = elem.getScrollTop();
                if (scrollTop <= 0) elem.setScrollTop(1);
                else if (scrollTop + elem.getOffsetHeight() >= elem.getScrollHeight()) {
                    elem.setScrollTop(elem.getScrollHeight() - elem.getOffsetHeight() - 1);
                }

                // Prevent the screen from shifting up and down if there is no overflow content to scroll.
                // https://github.com/joelambert/ScrollFix/pull/6
                // https://github.com/albertogasparin/ScrollFix/commit/d6200f040b30f6aab6fe9cec4773ab36613af595
                preventNativeScrolling = elem.getScrollHeight() <= elem.getOffsetHeight();
                break;

            case Event.ONTOUCHMOVE:
                if (preventNativeScrolling ||
                        !"touch".equals(self.getElement().getStyle().getProperty(DOMConstants.INSTANCE.getOverflowScrollingPropertyName())))
                {
                    event.preventDefault();
                }
                break;
        }
    }
}
