package com.smartgwt.mobile.client.widgets;

import java.util.Date;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.Widget;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants;

class ScrollablePanelImplNonNative extends ScrollablePanelImpl {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.scrollableCSS().getText());
    }

    private enum Direction {
        north,
        south
    };

    private boolean active = false, scrolled = false, scrollable = false, scrolling = false, scrollbarInitialized = false, mouseWheelActive = false;
    private int height, scrollbarHeight, startY = 0, lastY = 0, totalY = 0, wrapperHeight = 0;
    private long startTime = 0L, lastMouseWheelEventTime;
    private float barProp, deceleration = 0.001F;
    private Direction direction;
    private Event startEvent;
    private Timer endTimer = null;

    // The thin area at the very right of the ScrollablePanel that spans top to bottom.
    private Canvas scrollbar;

    // The visual indicator within the scrollbar of the current scroll position (the dark gray
    // rounded rectangle).
    private Canvas indicator;

    @Override
    void init(ScrollablePanel self) {
        scrollbar = new Canvas(Document.get().createDivElement());
        scrollbar.setClassName("sc-scrollable-bar");
        indicator = new Canvas(Document.get().createDivElement());
        indicator.setClassName("sc-scrollable-bar-indicator");
        scrollbar.addChild(indicator);
        self.add(scrollbar, self.getElement());
        self.sinkEvents(Event.GESTUREEVENTS | Event.TOUCHEVENTS | Event.MOUSEEVENTS | Event.FOCUSEVENTS | Event.ONMOUSEWHEEL | Event.ONSCROLL | Event.KEYEVENTS);
    }

    private void calculateScrollBarHeights(ScrollablePanel self) {
        scrollbarHeight = self.getElement().getClientHeight();
        int indicatorHeight = Math.round(((float)scrollbarHeight) * scrollbarHeight / height);
        indicator.getElement().getStyle().setHeight(indicatorHeight, Unit.PX);
        int barMaxScroll = scrollbarHeight - indicatorHeight - 4;
        final int maxScrollTop = height - scrollbarHeight;
        barProp = ((float)barMaxScroll) / maxScrollTop;
    }

    private int getIndicatorPos() {
        return Math.round(barProp * -totalY);
    }

    @Override
    void onBrowserEvent(ScrollablePanel self, Event event) {
        self.superOnBrowserEvent(event);
        switch (event.getTypeInt()) {
            case Event.ONMOUSEWHEEL:
                if (!mouseWheelActive) {
                    mouseWheelActive = true;
                    onStart(self, event);
                } else {
                    onMouseWheelMove(self, event);
                }
                break;
            case Event.ONMOUSEDOWN:
                if (!self.touched) {
                    onStart(self, event);
                }
                break;
            case Event.ONTOUCHSTART:
            case Event.ONGESTURESTART:
                self.touched = true;
                onStart(self, event);
                break;
            case Event.ONMOUSEMOVE:
                if (!self.touched) {
                    onMove(self, event);
                }
                break;
            case Event.ONTOUCHMOVE:
            case Event.ONGESTURECHANGE:
            case Event.ONSCROLL:
                onMove(self, event);
                break;
            case Event.ONMOUSEUP:
                if (!self.touched) {
                    onEnd(self, event);
                }
                break;
            case Event.ONTOUCHEND:
            case Event.ONTOUCHCANCEL:
            case Event.ONGESTUREEND:
                onEnd(self, event);
                break;
        }
    }

    private void onStart(ScrollablePanel self, Event event) {
        if (!self.isEnabled()) return;
        if (!active) {
            self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransitionShorthandPropertyName(), "none");
            lastY = (event.getTouches() != null && event.getTouches().length() > 0) ? event.getTouches().get(0).getClientY(): event.getClientY();
            height = self._getInnerElement().getOffsetHeight();
            final Widget parentWidget = self.getParent();
            if (parentWidget == null) {
                wrapperHeight = Window.getClientHeight();
            } else {
                wrapperHeight = parentWidget.getElement().getClientHeight();
            }
            if (!scrollbarInitialized) {
                calculateScrollBarHeights(self);
            }
            if (wrapperHeight < height) {
                active = true;
                scrollable = true;
                startEvent = event;
                startTime = new Date().getTime();
                int computedY = self.getComputedY(self._getInnerElement());
                if (computedY != totalY) {
                    totalY = self.getComputedY(self._getInnerElement());
                    if (Canvas.isAndroid()) {
                        self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + totalY + "px)");
                    } else {
                        self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + totalY + "px, 0px)");
                    }
                    final int indicatorPos = getIndicatorPos();
                    if (Canvas.isAndroid()) {
                        indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + indicatorPos + "px)");
                    } else {
                        indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + indicatorPos + "px, 0px)");
                    }
                    scrollbar.getElement().getStyle().setOpacity(0);
                    scrolled = true;
                    scrolling = true;
                }
            }
        }
    }

    private void onMove(ScrollablePanel self, Event event) {
        if (!self.isEnabled()) return;
        if (active) {
            if (scrollable) {
                scrolled = true;
                int Y = (event.getTouches() != null && event.getTouches().length() > 0) ? event.getTouches().get(0).getClientY(): event.getClientY();
                int deltaY = Y - lastY;
                lastY = Y;
                totalY += deltaY;
                direction = deltaY < 0 ? Direction.south : Direction.north;
                scrollbar.getElement().getStyle().setOpacity(1);
                if (Canvas.isAndroid()) {
                    self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + totalY + "px)");
                } else {
                    self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + totalY + "px, 0)");
                }
                final int indicatorPos = getIndicatorPos();
                if (Canvas.isAndroid()) {
                    indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + indicatorPos + "px)");
                } else {
                    indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + indicatorPos + "px, 0px)");
                }
            }
        }
    }

    private void onEnd(final ScrollablePanel self, Event event) {
        if (!self.isEnabled()) return;
        if (active && scrollable && scrolled && !scrolling) {
            boolean pastEnd = false;
            int duration = (int)(new Date().getTime() - startTime);
            int distance = Math.abs(totalY - startY);
            float speed = ((float)distance) / duration;
            int newDistance = Math.round((speed * speed) / (2 * deceleration));
            int newTime = Math.round(speed / deceleration);
            totalY += direction == Direction.north ? newDistance : -newDistance;
            if (totalY > 0) {
                totalY = 0;
                newTime = 500;
            } else if (Math.abs(totalY) > Math.abs(wrapperHeight-height)) {
                totalY = wrapperHeight - height;
                newTime = 500;
                pastEnd = true;
            }
            if (pastEnd || speed >= 0.3F) {
                final int indicatorPos = getIndicatorPos();
                self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransitionShorthandPropertyName(), DOMConstants.INSTANCE.getTransformPropertyNameForCSSText() + " " + newTime + "ms cubic-bezier(0.33,0.66,0.66,1)");
                indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransitionShorthandPropertyName(), DOMConstants.INSTANCE.getTransformPropertyNameForCSSText() + " " + newTime + "ms cubic-bezier(0.33,0.66,0.66,1)");
                if (Canvas.isAndroid()) {
                    self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + totalY + "px)");
                    indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + indicatorPos + "px)");
                } else {
                    self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + totalY + "px, 0px)");
                    indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + indicatorPos + "px, 0px)");
                }
            }
            endTimer = new Timer() {
                @Override
                public void run() {
                    if (endTimer != this) return;
                    self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransitionShorthandPropertyName(), "none 0s");
                    indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransitionShorthandPropertyName(), "none 0s");
                    scrollbar.getElement().getStyle().setOpacity(0);
                }
            };
            endTimer.schedule(newTime + 20);
        }
        active = false;
        scrolled = false;
        scrolling = false;
        scrollable = false;
        startEvent = null;
        startY = totalY;
        startTime = 0;
    }

    private void onMouseWheelMove(final ScrollablePanel self, final Event event) {
        if (!self.isEnabled()) return;
        if (active) {
            if (scrollable) {
                scrolled = true;
                int deltaY = self.getWheelDelta(event);
                lastY += deltaY;
                totalY += deltaY;
                direction = deltaY < 0 ? Direction.south : Direction.north;
                scrollbar.getElement().getStyle().setOpacity(1);
                final int indicatorPos = getIndicatorPos();
                if (Canvas.isAndroid()) {
                    self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + totalY + "px)");
                    indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate(0px, " + indicatorPos + "px)");
                } else {
                    self._getInnerElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + totalY + "px, 0px)");
                    indicator.getElement().getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translate3d(0px, " + indicatorPos + "px, 0px)");
                }
                lastMouseWheelEventTime = new Date().getTime();
                new Timer() {
                    @Override
                    public void run() {
                        if (mouseWheelActive) {
                            long currentTime = new Date().getTime();
                            final int mouseWheelDeltaTime = new Long(currentTime - lastMouseWheelEventTime).intValue();
                            if (mouseWheelDeltaTime > 80) {
                                onEnd(self, event);
                                mouseWheelActive = false;
                            }
                        }
                    }
                }.schedule(80);
            }
        }
    }
}
