package com.smartgwt.mobile.client.widgets.form.fields;

import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;

import com.google.gwt.core.client.JsDate;
import com.google.gwt.i18n.shared.DateTimeFormat;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.widgets.Picker;
import com.smartgwt.mobile.client.internal.widgets.events.ValuesSelectedEvent;
import com.smartgwt.mobile.client.types.DateItemSelectorFormat;
import com.smartgwt.mobile.client.widgets.layout.Layout;

public class DateItem extends SelectItem {

    private static final DateItemSelectorFormat DEFAULT_DATE_ITEM_SELECTOR_FORMAT = DateItemSelectorFormat.MONTH_DAY_YEAR;
    private static final LinkedHashMap<Integer, String> DEFAULT_MONTHS_MAP = new LinkedHashMap<Integer, String>() {{
        put(0, "January");
        put(1, "February");
        put(2, "March");
        put(3, "April");
        put(4, "May");
        put(5, "June");
        put(6, "July");
        put(7, "August");
        put(8, "September");
        put(9, "October");
        put(10, "November");
        put(11, "December");
    }};
    private static final LinkedHashMap<Integer, String> DEFAULT_DATES_MAP = new LinkedHashMap<Integer, String>() {{
        for (int i = 1; i <= 31; ++i) {
            put(i, Integer.toString(i));
        }
    }};

    private static boolean isValidDateRange(Date startDate, Date endDate) {
        if (startDate == null || endDate == null) return true;
        final JsDate startJsD = JsDate.create((double)startDate.getTime()),
                endJsD = JsDate.create((double)endDate.getTime());
        if (startJsD.getFullYear() < endJsD.getFullYear()) return true;
        else if (startJsD.getFullYear() == endJsD.getFullYear()) {
            if (startJsD.getMonth() < endJsD.getMonth()) return true;
            else if (startJsD.getMonth() == endJsD.getMonth()) {
                return (startJsD.getDate() <= endJsD.getDate());
            }
        }
        return false;
    }

    private Date startDate, endDate;
    private boolean useTextField = false;
    private DateItemSelectorFormat dateItemSelectorFormat;
    private DateTimeFormat formatter;

	public DateItem(String name) {
		super(name);
        setDateItemSelectorFormat(DEFAULT_DATE_ITEM_SELECTOR_FORMAT);
	}

    public DateItem(String name, String title, String placeholder) {
        this(name);
        setTitle(title);
        setPlaceholder(placeholder);
    }

    private LinkedHashMap<Integer, String> createDatesMap() {
        assert isValidDateRange(startDate, endDate);
        LinkedHashMap<Integer, String> ret = DEFAULT_DATES_MAP;
        if (startDate != null && endDate != null) {
            final JsDate startJsD = JsDate.create((double)startDate.getTime()),
                    endJsD = JsDate.create((double)endDate.getTime());
            if (startJsD.getFullYear() == endJsD.getFullYear() &&
                    startJsD.getMonth() == endJsD.getMonth())
            {
                ret = new LinkedHashMap<Integer, String>();
                final int endI = endJsD.getDate();
                for (int i = startJsD.getDate(); i <= endI; ++i) {
                    ret.put(i, Integer.toString(i));
                }
            }
        }
        return ret;
    }

    private LinkedHashMap<Integer, String> createMonthsMap() {
        assert isValidDateRange(startDate, endDate);
        LinkedHashMap<Integer, String> ret = DEFAULT_MONTHS_MAP;
        if (startDate != null && endDate != null) {
            final JsDate startJsD = JsDate.create((double)startDate.getTime()),
                    endJsD = JsDate.create((double)endDate.getTime());
            if (startJsD.getFullYear() == endJsD.getFullYear()) {
                ret = new LinkedHashMap<Integer, String>();
                final int endI = endJsD.getMonth();
                for (int i = startJsD.getMonth(); i <= endI; ++i) {
                    ret.put(i, DEFAULT_MONTHS_MAP.get(i));
                }
            }
        }
        return ret;
    }

    private LinkedHashMap<Integer, String> createYearsMap() {
        assert isValidDateRange(startDate, endDate);
        final JsDate startJsD, endJsD;
        if (startDate != null && endDate != null) {
            startJsD = JsDate.create((double)startDate.getTime());
            endJsD = JsDate.create((double)endDate.getTime());
        } else {
            startJsD = JsDate.create();
            startJsD.setFullYear(startJsD.getFullYear() - 5);
            endJsD = JsDate.create(startJsD.getTime());
            endJsD.setFullYear(endJsD.getFullYear() + 10);
        }
        final LinkedHashMap<Integer, String> ret = new LinkedHashMap<Integer, String>();
        final int endI = endJsD.getFullYear();
        for (int i = startJsD.getFullYear(); i <= endI; ++i) {
            ret.put(i, Integer.toString(i));
        }
        return ret;
    }

    private Integer[] getPickerValues(Date d) {
        if (d == null) return null;
        final JsDate jsD = JsDate.create((double)d.getTime());
        final Integer[] values;
        switch (dateItemSelectorFormat) {
            case DAY_MONTH:
                values = new Integer[] {
                        jsD.getDate(),
                        jsD.getMonth()
                    };
                break;
            case DAY_MONTH_YEAR:
                values = new Integer[] {
                        jsD.getDate(),
                        jsD.getMonth(),
                        jsD.getFullYear()
                    };
                break;
            case MONTH_DAY:
                values = new Integer[] {
                        jsD.getMonth(),
                        jsD.getDate()
                    };
                break;
            case MONTH_DAY_YEAR:
                values = new Integer[] {
                        jsD.getMonth(),
                        jsD.getDate(),
                        jsD.getFullYear()
                    };
                break;
            case MONTH_YEAR:
                values = new Integer[] {
                        jsD.getMonth(),
                        jsD.getFullYear()
                    };
                break;
            case YEAR_MONTH:
                values = new Integer[] {
                        jsD.getFullYear(),
                        jsD.getMonth()
                    };
                break;
            case YEAR_MONTH_DAY:
                values = new Integer[] {
                        jsD.getFullYear(),
                        jsD.getMonth(),
                        jsD.getDate()
                    };
                break;
            default:
                assert false;
                return null;
        }
        return values;
    }

    @SGWTInternal
    protected Picker _createPicker(Layout pickerParent, Layout pickerScreenCoverParent) {
        final List<LinkedHashMap<Integer, String>> valueMaps = new ArrayList<LinkedHashMap<Integer, String>>();
        switch (dateItemSelectorFormat) {
            case DAY_MONTH:
                valueMaps.add(createDatesMap());
                valueMaps.add(createMonthsMap());
                break;
            case DAY_MONTH_YEAR:
                valueMaps.add(createDatesMap());
                valueMaps.add(createMonthsMap());
                valueMaps.add(createYearsMap());
                break;
            case MONTH_DAY:
                valueMaps.add(createMonthsMap());
                valueMaps.add(createDatesMap());
                break;
            case MONTH_DAY_YEAR:
                valueMaps.add(createMonthsMap());
                valueMaps.add(createDatesMap());
                valueMaps.add(createYearsMap());
                break;
            case MONTH_YEAR:
                valueMaps.add(createMonthsMap());
                valueMaps.add(createYearsMap());
                break;
            case YEAR_MONTH:
                valueMaps.add(createYearsMap());
                valueMaps.add(createMonthsMap());
                break;
            case YEAR_MONTH_DAY:
                valueMaps.add(createYearsMap());
                valueMaps.add(createMonthsMap());
                valueMaps.add(createDatesMap());
                break;
        }
        final Picker picker = new Picker(pickerParent, pickerScreenCoverParent, valueMaps.toArray(new LinkedHashMap[valueMaps.size()]));
        picker.setSelectedValues(getPickerValues(getValueAsDate()));
        return picker;
    }

    @Override
    public void _onValuesSelected(ValuesSelectedEvent event) {
        final Object[] selectedValues = event.getValues();
        if (selectedValues == null) {
            setValue(null);
            return;
        } else {
            int dateIndex = -1, monthIndex = -1, fullYearIndex = -1;
            switch (dateItemSelectorFormat) {
                case DAY_MONTH:
                    assert selectedValues.length == 2;
                    dateIndex = 0;
                    monthIndex = 1;
                    break;
                case DAY_MONTH_YEAR:
                    assert selectedValues.length == 3;
                    dateIndex = 0;
                    monthIndex = 1;
                    fullYearIndex = 2;
                    break;
                case MONTH_DAY:
                    assert selectedValues.length == 2;
                    monthIndex = 0;
                    dateIndex = 1;
                    break;
                case MONTH_DAY_YEAR:
                    assert selectedValues.length == 3;
                    monthIndex = 0;
                    dateIndex = 1;
                    fullYearIndex = 2;
                    break;
                case MONTH_YEAR:
                    assert selectedValues.length == 2;
                    monthIndex = 0;
                    fullYearIndex = 1;
                    break;
                case YEAR_MONTH:
                    assert selectedValues.length == 2;
                    fullYearIndex = 0;
                    monthIndex = 1;
                    break;
                case YEAR_MONTH_DAY:
                    assert selectedValues.length == 3;
                    fullYearIndex = 0;
                    monthIndex = 1;
                    dateIndex = 2;
                    break;
            }

            // JavaScript has a gotcha where if you set the month of a Date object, the value of
            // getMonth() may not be what you just set it to! For example:
            // var d = new Date(1346450400000); // August 31, 2012 at 3:00 PM PDT
            // d.setMonth(4 - 1); // April
            // d.getMonth() == (5 - 1); // The month is now May; d is May 1, 2012 at 3:00 PM PDT.
            //
            // We therefore set the full year, month, and date all at once.
            final JsDate jsD = startDate == null ? JsDate.create() : JsDate.create((double)startDate.getTime());
            int fullYear = jsD.getFullYear(),
                    month = jsD.getMonth(),
                    date = jsD.getDate();
            if (dateIndex != -1) {
                final Integer i = (Integer)selectedValues[dateIndex];
                if (i != null) date = i.intValue();
            }
            if (monthIndex != -1) {
                final Integer i = (Integer)selectedValues[monthIndex];
                if (i != null) month = i.intValue();
            }
            if (fullYearIndex != -1) {
                final Integer i = (Integer)selectedValues[fullYearIndex];
                if (i != null) fullYear = i.intValue();
            }
            jsD.setFullYear(fullYear, month, date);

            setValue(new Date((long)jsD.getTime()));
        }
    }

    @SGWTInternal
    protected Object _mapValueToDisplay(Object value) {
        assert formatter != null;
        if (value == null) return "";
        else if (value instanceof Date) return formatter.format((Date)value);
        return value;
    }

    public final Date getStartDate() {
        return startDate;
    }

    public void setStartDate(Date startDate) {
        if (!isValidDateRange(startDate, endDate)) throw new IllegalArgumentException("startDate");
        this.startDate = startDate;
    }

    public final Date getEndDate() {
        return endDate;
    }

    public void setEndDate(Date endDate) {
        if (!isValidDateRange(startDate, endDate)) throw new IllegalArgumentException("endDate");
        this.endDate = endDate;
    }

    public final boolean getUseTextField() {
        return useTextField;
    }

    public void setUseTextField(boolean useTextField) {
        this.useTextField = useTextField;
    }

    public final DateItemSelectorFormat getDateItemSelectorFormat() {
        return dateItemSelectorFormat;
    }

    public void setDateItemSelectorFormat(DateItemSelectorFormat dateItemSelectorFormat) {
        if (dateItemSelectorFormat == null) dateItemSelectorFormat = DEFAULT_DATE_ITEM_SELECTOR_FORMAT;
        this.dateItemSelectorFormat = dateItemSelectorFormat;
        switch (dateItemSelectorFormat) {
            case DAY_MONTH:
                formatter = DateTimeFormat.getFormat("dd-MM");
                break;
            case DAY_MONTH_YEAR:
                formatter = DateTimeFormat.getFormat("dd-MM-yyyy");
                break;
            case MONTH_DAY:
                formatter = DateTimeFormat.getFormat("MM-dd");
                break;
            case MONTH_DAY_YEAR:
                formatter = DateTimeFormat.getFormat("dd-MM-yyyy");
                break;
            case MONTH_YEAR:
                formatter = DateTimeFormat.getFormat("MM-yyyy");
                break;
            case YEAR_MONTH:
                formatter = DateTimeFormat.getFormat("yyyy-MM");
                break;
            case YEAR_MONTH_DAY:
                formatter = DateTimeFormat.getFormat("yyyy-MM-dd");
                break;
        }
    }

    public Date getValueAsDate() {
        assert formatter != null;
        if (_value == null) return null;
        else if (_value instanceof Date) return ((Date)_value);
        else {
            try {
                return formatter.parse(getValueAsString());
            } catch (IllegalArgumentException ex) {
                return null;
            }
        }
    }

    @Override
    public boolean validate() {
        Object obj = getValue();
		return obj instanceof Date;
	}
}
