package com.smartgwt.mobile.client.widgets.form.fields;

import java.util.LinkedHashMap;

import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.InputElement;
import com.google.gwt.event.dom.client.HasKeyPressHandlers;
import com.google.gwt.event.dom.client.KeyPressEvent;
import com.google.gwt.event.dom.client.KeyPressHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.data.Criterion;
import com.smartgwt.mobile.client.internal.EventUtil;
import com.smartgwt.mobile.client.internal.widgets.Picker;
import com.smartgwt.mobile.client.internal.widgets.events.HasValuesSelectedHandlers;
import com.smartgwt.mobile.client.internal.widgets.events.PickerHiddenEvent;
import com.smartgwt.mobile.client.internal.widgets.events.PickerHiddenHandler;
import com.smartgwt.mobile.client.internal.widgets.events.ValuesSelectedEvent;
import com.smartgwt.mobile.client.internal.widgets.events.ValuesSelectedHandler;
import com.smartgwt.mobile.client.internal.widgets.form.fields.events.InputHandler;
import com.smartgwt.mobile.client.types.FormItemType;
import com.smartgwt.mobile.client.types.OperatorId;
import com.smartgwt.mobile.client.widgets.Canvas;
import com.smartgwt.mobile.client.widgets.form.DynamicForm;
import com.smartgwt.mobile.client.widgets.form.FormItemValueFormatter;
import com.smartgwt.mobile.client.widgets.form.events.ItemChangedEvent;
import com.smartgwt.mobile.client.widgets.form.fields.events.BlurEvent;
import com.smartgwt.mobile.client.widgets.form.fields.events.BlurHandler;
import com.smartgwt.mobile.client.widgets.form.fields.events.ChangedEvent;
import com.smartgwt.mobile.client.widgets.form.fields.events.ChangedHandler;
import com.smartgwt.mobile.client.widgets.form.fields.events.HasBlurHandlers;
import com.smartgwt.mobile.client.widgets.form.fields.events.HasChangedHandlers;
import com.smartgwt.mobile.client.widgets.layout.Layout;

public abstract class FormItem extends Canvas implements ValuesSelectedHandler, HasBlurHandlers, HasChangedHandlers, HasKeyPressHandlers, HasValuesSelectedHandlers {

    static boolean inputActive = false;

    private DynamicForm form;
    private Layout pickerParent, pickerScreenCoverParent;
    @SGWTInternal
    protected Picker _picker;
    private LinkedHashMap<?, String> valueMap = null;

    @SGWTInternal
    Object _value;
    private Object changeValue;
    private Object defaultValue = null;
    private String name;
    private String title;
    private String placeholder;
    private boolean canEdit = true;
    private boolean showTitle = true;
    FormItemType type = FormItemType.TEXT;
    private boolean required = false, shouldShowErrorText = false, shouldShowErrorIcon = true;
    @SGWTInternal
    boolean _changeOnKeypress = true;

	protected FormItem(String name, Element input) {
		this.name = name;
        setElement(input);
        getElement().setId(name);
        setClassName("sc-layout-flex0");
        sinkEvents(Event.ONCLICK | Event.ONMOUSEDOWN | Event.ONMOUSEMOVE | Event.ONMOUSEUP | Event.TOUCHEVENTS | Event.ONFOCUS | Event.ONBLUR);
    }

    @SGWTInternal
    native HandlerRegistration _addInputHandler(Element element, final InputHandler handler) /*-{
        var self = this;
        var listener = function (event) {
            handler.@com.smartgwt.mobile.client.internal.widgets.form.fields.events.InputHandler::_onInput()();
        };
        element.addEventListener("input", listener, false);
        return new @com.smartgwt.mobile.client.internal.widgets.events.NativeHandlerRegistration::new(Lcom/google/gwt/dom/client/Element;Ljava/lang/String;Lcom/google/gwt/core/client/JavaScriptObject;Z)(element, "input", listener, false);
    }-*/;

    @SGWTInternal
    void _handleInput() {
        if (_changeOnKeypress) {
            _updateValue();
        } // else TODO length 
    }

    public final DynamicForm getForm() {
        return form;
    }

    @SGWTInternal
    public void _setForm(DynamicForm form) {
        this.form = form;
    }

	public Criterion getCriterion() {
        return new Criterion(name, OperatorId.EQUALS, _value);
    }

    public void setTitle(String title) {
		this.title = title;
	}

	public String getTitle() {
		return title;
	}

    public final boolean getShowTitle() {
        return showTitle;
    }

    public void setShowTitle(Boolean showTitle) {
        this.showTitle = showTitle == null ? true : showTitle.booleanValue();
    }

	public void setName(String name) {
		this.name = name;
	}
	
	public String getName() {
		return name;
	}

    public final FormItemType getType() {
        return type;
    }

	public final String getPlaceholder() {
        return placeholder;
    }

    public void setPlaceholder(String placeholder) {
        this.placeholder = placeholder;
        getElement().setAttribute("placeholder", this.placeholder);
    }

    public final boolean getCanEdit() {
        return canEdit;
    }

    public void setCanEdit(Boolean canEdit) {
        this.canEdit = canEdit == null ? true : canEdit.booleanValue();
    }

    public final boolean isRequired() {
		return required;
	}

	public void setRequired(boolean required) {
		this.required = required;
	}

    public Object getValue() {
        return _value;
        // TODO form.getSavedItemValue() if `value` is null.
    }

    public void setValue(Object newValue) {
        setValue(newValue, false, false);
    }

    @SGWTInternal
    protected Object _mapValueToDisplay(Object value) {
        if (valueMap != null) return valueMap.get(value);
        return value;
    }

    protected Object setValue(Object newValue, boolean allowNullValue, boolean timeCritical) {

        // TODO _setValueCalled, resetCursor, _pendingUpdate

        boolean isDefault = false;
        if (newValue == null && ! allowNullValue) {
            Object defaultVal = getDefaultValue();
            if (defaultVal != null) {
                isDefault = true;
                newValue = defaultVal;
            }
        }

        // TODO length

        saveValue(newValue, isDefault);
        
        _handleChanged(newValue);

        // TODO shouldFetchMissingValue

        Object displayValue = _mapValueToDisplay(newValue);
        _setElementValue(displayValue, newValue);

        // TODO resetCursor

        return newValue;
    }

    public final Object getDefaultValue() {
        // TODO defaultDynamicValue

        return defaultValue;
    }

    @SGWTInternal
    void _updateValue() {
        // TODO hasElement()

        Object newValue = _getElementValue();
        _updateValue(newValue);
    }

    @SGWTInternal
    void _updateValue(Object newValue) {
        // TODO _pendingCompletion, mapDisplayToValue

        _storeValue(newValue);
    }

    private boolean compareValues(Object value1, Object value2) {
        return DynamicForm._compareValues(value1, value2);
    }

    @SGWTInternal
    boolean _storeValue(Object newValue) {
        if (compareValues(newValue, _value)) {
            //SC.logWarn("FI._updateValue: not saving, value unchanged: " + value);
            return true;
        }

        // TODO changingValue

        boolean returnVal = handleChange(newValue, _value);

        // Ensure we have the latest value
        newValue = changeValue;

        // TODO updateAppearance()

        if (! compareValues(newValue, _value)) saveValue(newValue, false);
        changeValue = null;

        _handleChanged(_value);
        return returnVal;
    }

    private void saveValue(Object value, boolean isDefault) {
        this._value = value;

        // TODO _setToDefault, form.clearItemValue() or form.saveItemValue()
    }

    private boolean handleChange(Object newValue, Object value) {
        // TODO
        changeValue = newValue;
        return true;
    }

    @SGWTInternal
    void _handleChanged(Object value) {
        // TODO form.rulesEngine

        ChangedEvent.fire(this, form, this, value);
        if (form != null) {
            ItemChangedEvent.fire(form, form, this, value);

            // TODO checkForImplicitSave()
        }
    }

    @SGWTInternal
    Object _getElementValue() {
        return getElement().<InputElement>cast().getValue();
    }

    @SGWTInternal
    void _setElementValue(Object displayValue, Object newValue) {
        String str = displayValue == null ? null : displayValue.toString();
        if (str == null) str = "";
        getElement().<InputElement>cast().setValue(str);
    }

    @SGWTInternal
    protected final LinkedHashMap<?, String> _getValueMap() {
        return valueMap;
    }

    public final void setValueMap(String... values) {
        final LinkedHashMap<String, String> valueMap = new LinkedHashMap<String, String>();
        if (values != null) {
            for (final String value : values) {
                valueMap.put(value,  value);
            }
        }
        setValueMap(valueMap);
    }

    public void setValueMap(LinkedHashMap<?, String> valueMap) {
        this.valueMap = valueMap;
        setValue(_value);
    }

    @Override
    public void onBrowserEvent(Event event) {
        super.onBrowserEvent(event);

        Element target = EventUtil.getTargetElem(event);
        if (target != null && getElement().isOrHasChild(target)) {
            int type = event.getTypeInt();
            switch (type) {
                case Event.ONMOUSEDOWN:
                    onStart(event);
                    break;
                case Event.ONTOUCHSTART:
                    touched = true;
                    onStart(event);
                    break;
                case Event.ONMOUSEUP:
                    onEnd(event);
                    break;
                case Event.ONTOUCHEND:
                case Event.ONTOUCHCANCEL:
                    onEnd(event);
                    break;
                case Event.ONFOCUS:
                    inputActive = true;
                    break;
                case Event.ONBLUR:
                    onBlur(event);
                    break;
            }
        }
    }

    protected void onBlur(final Event event) {
		if (!isEnabled()) {
			return;
		}
		InputElement input = getElement().cast();
    	setValue(input.getValue());  
    	inputActive = false;
        if (canEdit) BlurEvent.fire(this, form, this);

        new Timer() {
            public void run() {
                if(!inputActive) {
                    Canvas._maybeHideAddressBar();
                }
            }
        }.schedule(200);
    }

    protected void onStart(Event event) {
        if (!isEnabled()) {
            return;
        }
    }

    protected void onEnd(Event event) {
        if (canEdit && pickerParent != null && pickerScreenCoverParent != null) {
            showPicker();
        }
    }

    protected void onChanged(Event event) {
        final DynamicForm form = this.form;
        new Timer() {
            @Override
            public void run() {
                ChangedEvent.fire(FormItem.this, form, FormItem.this, _value);
            }
        }.schedule(20);
    }

    public void enablePicker(Layout parent, Layout screenCoverParent) {
        pickerParent = parent;
        pickerScreenCoverParent = screenCoverParent;
    }

    @SGWTInternal
    protected Picker _createPicker(Layout pickerParent, Layout pickerScreenCoverParent) {
        return new Picker(pickerParent, pickerScreenCoverParent, valueMap);
    }

    public void showPicker() {
        _picker = _createPicker(pickerParent, pickerScreenCoverParent);
        final HandlerRegistration valuesSelectedRegistration = _picker._addValuesSelectedHandler(this);
        new PickerHiddenHandler() {

            private HandlerRegistration pickerHiddenRegistration = _picker._addPickerHiddenHandler(this);

            @Override
            public void _onPickerHidden(PickerHiddenEvent event) {
                pickerHiddenRegistration.removeHandler();
                valuesSelectedRegistration.removeHandler();
                _picker.destroy();
                _picker = null;

                BlurEvent.fire(FormItem.this, form, FormItem.this);
            }
        };
        _picker.show();
    }

    public void setValueFormatter(FormItemValueFormatter formatter) {
        
    }

    public abstract boolean validate();

    @SGWTInternal
	public static FormItem _create(String name, String type) {
		if(type.equals("tel")) {
			return new PhoneItem(name);
		} else if(type.equals("url")) {
			return new UrlItem(name);
		} else if(type.equals("email")) {
			return new EmailItem(name);
		} else if(type.equals("search")) {
			return new SearchItem(name);
		} else if(type.equals("datetime")) {
			return new DateItem(name);
		} else if(type.equals("checkbox")) {
			return new CheckboxItem(name);
		}
		return new TextItem(name);
	}

    @Override
    public void _onValuesSelected(ValuesSelectedEvent event) {
        final Object[] values = event.getValues();
        if (values == null) setValue(null);
        else {
            if (values.length == 1) {
                setValue(values[0]);
            } else {
                setValue(values);
            }
        }
    }

    @Override
    public HandlerRegistration addBlurHandler(BlurHandler handler) {
        return addHandler(handler, BlurEvent.getType());
    }

    @Override
    public HandlerRegistration addChangedHandler(ChangedHandler handler) {
        return addHandler(handler, ChangedEvent.getType());
    }

    @Override
    public HandlerRegistration addKeyPressHandler(KeyPressHandler handler) {
        return addHandler(handler, KeyPressEvent.getType());
    }
    
    @Override
    public HandlerRegistration _addValuesSelectedHandler(ValuesSelectedHandler handler) {
        return addHandler(handler, com.smartgwt.mobile.client.internal.widgets.events.ValuesSelectedEvent.getType());
    }
}
