package com.smartgwt.mobile.client.widgets.form.fields;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.google.gwt.dom.client.Element;
import com.google.gwt.user.client.Event;
import com.smartgwt.mobile.client.data.Criteria;
import com.smartgwt.mobile.client.data.DataSource;
import com.smartgwt.mobile.client.data.DataSourceField;
import com.smartgwt.mobile.client.data.Record;
import com.smartgwt.mobile.client.data.ResultSet;
import com.smartgwt.mobile.client.data.events.DataArrivedEvent;
import com.smartgwt.mobile.client.data.events.DataArrivedHandler;
import com.smartgwt.mobile.client.widgets.grid.ListGrid;
import com.smartgwt.mobile.client.widgets.grid.ListGridField;

public class SelectItem extends TextItem implements PickList {
    Criteria pickListCriteria;
    ListGrid pickListProperties;
    List<ListGridField> pickListFields;
    DataSource optionDataSource;
    String displayField, valueField;
    Boolean defaultToFirstOption = true;

    public SelectItem(String name) {
        super(name);
    }

    public SelectItem(String name, String title) {
        this(name);
        setTitle(title);
    }

    public SelectItem(String name, String title, String placeholder) {
        this(name, title);
        setPlaceholder(placeholder);
    }

    public void fetchData() {
        DataSource ds = optionDataSource;
        if (ds == null) {
            return;
        }
        final ResultSet rs = new ResultSet(ds);
        if(pickListCriteria != null) {
            rs.setCriteria(pickListCriteria);
        }
        rs.addDataArrivedHandler(new DataArrivedHandler() {
            @Override
            public void onDataArrived(DataArrivedEvent event) {
                for (int i = event.getStartRow(); i < event.getEndRow(); ++i) {
                    Record record = rs.get(i);
                    String[] pickList = record.getAttributeAsStringArray(displayField);
                    setValueMap(pickList);
                }
            }
        });
    }
        
    public ListGrid getPickListProperties() {
        return pickListProperties;
    }

    public void setPickListProperties(ListGrid pickListProperties) {
        this.pickListProperties = pickListProperties;
    }

    public Boolean getDefaultToFirstOption() {
        return defaultToFirstOption;
    }

    public void setDefaultToFirstOption(Boolean defaultToFirstOption) {
        this.defaultToFirstOption = defaultToFirstOption;
        final LinkedHashMap<?, String> valueMap = _getValueMap();
        if(defaultToFirstOption && valueMap != null && !valueMap.isEmpty()) {
            final Iterator<?> it = valueMap.entrySet().iterator();
            assert it.hasNext();
            setValue(((Map.Entry<?, ?>)it.next()).getKey());
        }
    }
  
    @Override
    public void onBrowserEvent(Event event) {
        event.preventDefault();
        event.stopPropagation();
        Element target = event.getEventTarget().cast();
        if (target != null && getElement().isOrHasChild(target)) {
            int type = event.getTypeInt();
            switch (type) {
                case Event.ONMOUSEDOWN:
                    onStart(event);
                    break;
                case Event.ONTOUCHSTART:
                    touched = true;
                    onStart(event);
                    break;
                case Event.ONMOUSEUP:
                    onEnd(event);
                    break;
                case Event.ONTOUCHEND:
                case Event.ONTOUCHCANCEL:
                    onEnd(event);
                    break;
                case Event.ONFOCUS:
                    inputActive = true;
                    break;
                case Event.ONBLUR:
                    onBlur(event);
                    break;
            }
        }
    }

    @Override
    public Record[] getClientPickListData() {
        // TODO Auto-generated method stub
        return null;
    }
    
    public void setPickListFields(ListGridField... pickListFields) {
        for(int i = 0; i < pickListFields.length; ++i) {
            this.pickListFields.add(pickListFields[i]);
        }
    }

    @Override
    public String getDisplayField() {
        if(displayField != null) {
            return displayField;
        } else if(valueField != null) {
            return valueField;
        }
        return getName();
    }

    @Override
    public DataSource getOptionDataSource() {
        return optionDataSource;
    }

    @Override
    public String getValueField() {
        return valueField;
    }

    @Override
    public void setDisplayField(String displayField) {
        this.displayField = displayField;
    }

    @Override
    public void setOptionDataSource(DataSource ds) {
        this.optionDataSource = ds;
    }

    @Override
    public void setValueField(String valueField) {
        this.valueField = valueField;
    }

    @Override
    public Criteria getPickListCriteria() {
        return pickListCriteria;
    }

    @Override
    public void setPickListCriteria(Criteria criteria) {
        this.pickListCriteria = criteria;        
    }

    @Override
    public String getDisplayFieldName() {
        if(displayField != null) {
            return displayField;
        }
        if(optionDataSource != null) {
            String valueFieldName = getValueFieldName();
            DataSourceField dataSourceField = optionDataSource.getField(valueFieldName);
            if(dataSourceField != null && dataSourceField.isHidden()) {
                return optionDataSource.getTitleField();
            }
        }
        return null;
    }

    @Override
    public String getValueFieldName() {
        if(valueField != null) {
            return valueField;
        }
        String name = getName();
        if(name != null) {
            return name;
        }
        return "name";
    }

}
