package com.smartgwt.mobile.client.widgets.form.fields;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.SpanElement;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.types.FormItemType;
import com.smartgwt.mobile.client.widgets.Canvas;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants;

public class SwitchItem extends FormItem {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.switchItemCSS().getText());
    }

    private static final int THUMBX = 49, ONX = 77;
	private int offsetX, lastThumbX = 0, lastOnX = 0;
	private boolean active = false;
    private String thumbClassName = "sc-switchitem-thumb";
    private CheckboxItem checked;
	private SpanElement thumb;

    {
        type = FormItemType.BOOLEAN;
    }

    public SwitchItem(String name) {
        super(name, Document.get().createDivElement());
    	setClassName("sc-switchitem");
        getElement().addClassName(className);
        if(!Canvas.isAndroid()) {
            getElement().addClassName("shadow");
        }
        thumb = Document.get().createSpanElement();
        thumb.addClassName(thumbClassName);
        getElement().appendChild(thumb);
        checked = new CheckboxItem(name+"_internal");
        checked.setValue(false);
        getElement().appendChild(checked.getElement());
        setTransform(false);
        sinkEvents(Event.GESTUREEVENTS);
    }

    public SwitchItem(String name, String title) {
        this(name);
        setTitle(title);
    }

    private void setTransform(boolean transition) {
    	if(transition) {
            thumb.getStyle().setProperty(DOMConstants.INSTANCE.getTransitionShorthandPropertyName(), DOMConstants.INSTANCE.getTransformPropertyNameForCSSText() + " 125ms ease-in-out");
    	} else {
            thumb.getStyle().setProperty(DOMConstants.INSTANCE.getTransitionShorthandPropertyName(), "none 0s");
    	}
        thumb.getStyle().setProperty(DOMConstants.INSTANCE.getTransformPropertyName(), "translateX("+lastThumbX+"px)");
        getElement().getStyle().setProperty(DOMConstants.INSTANCE.getBackgroundSizePropertyName(), lastOnX + "px 28px, 100% 28px");
    }

    @Override
    public void onBrowserEvent(Event event) {
        if (!isEnabled()) {
            return;
        }
        Element target = event.getEventTarget().cast();
        if (target != null && getElement().isOrHasChild(target)) {
	        switch (event.getTypeInt()) {
	        	case Event.ONMOUSEDOWN:
	        	    if(!touched) {
	        	        onStart(event);
	        	    }
	        	    break;
	        	case Event.ONTOUCHSTART:
	        	case Event.ONGESTURESTART:
	        	    touched = true;
	        		onStart(event);
	        		break;
	        	case Event.ONMOUSEMOVE:
	        	    if(!touched) {
	        	        onMove(event);
	        	    }
	        	    break;
	        	case Event.ONTOUCHMOVE:
	        	case Event.ONGESTURECHANGE:
	        		onMove(event);
	        		break;
	            case Event.ONCLICK:
	            case Event.ONMOUSEUP:
	                if(!touched) {
	                    onEnd(event);                   
	                }
	                break;
	        	case Event.ONTOUCHEND:
	        	case Event.ONTOUCHCANCEL:
	        	case Event.ONGESTUREEND:
	        		onEnd(event);
	        		break;
	        }
        }
    }

    @Override
    public void onStart(Event event) {
    	if(!active) {
        	event.preventDefault();
        	event.stopPropagation();
            if (!getCanEdit()) return;
        	active = true;
        	lastThumbX = isChecked() ? 0 : THUMBX;
        	lastOnX = isChecked() ? ONX : 0;
        	offsetX = (event.getTypeInt() == Event.ONTOUCHSTART ? event.getTouches().get(0).getClientX() : event.getClientX());
    	}
    }

    protected void onMove(Event event) {
    	if(active) {
        	event.preventDefault();
            event.stopPropagation();
        	int newX = (event.getTypeInt() == Event.ONTOUCHMOVE ? event.getTouches().get(0).getClientX() : event.getClientX());
    		int deltaX = newX - offsetX;
    		offsetX = newX;
    		lastThumbX += deltaX;
    		lastOnX += deltaX;
    		if(lastOnX < 0) {
    		    lastOnX = 0;
    		}
    		if(lastOnX > ONX) {
    		    lastOnX = ONX;
    		}
    		if(lastThumbX < 0) {
    			lastThumbX = 0;
    		}
    		if(lastThumbX > THUMBX) {
    			lastThumbX = THUMBX;
    		}
    		setTransform(false);
    	}
    }

    protected void onEnd(final Event event) {
    	if(active) {
          	event.preventDefault();
            event.stopPropagation();
    		active = false;
    		setChecked(!isChecked());
            new Timer() {
                public void run() {
                    ClickEvent._fire(SwitchItem.this, event.getClientX(), event.getClientY());
                }
            }.schedule(1);
        }
    }

    public boolean isChecked() {
    	return checked.getValue() == Boolean.TRUE;
    }

    public void setChecked(boolean onoff) {
    	checked.setValue(onoff);
    	_value = onoff;
    	if(isChecked()) {
    	    lastThumbX = THUMBX;
    	    lastOnX = ONX;
    	} else {
    	    lastThumbX = 0;
    	    lastOnX = 0;
    	}
    	setTransform(true);
    }

    @Override
    Boolean _getElementValue() {
        return checked._getElementValue();
    }

    @Override
    void _setElementValue(Object displayValue, Object newValue) {
        boolean check = false;
        if (displayValue != null && displayValue instanceof Boolean) check = ((Boolean)displayValue).booleanValue();
        setChecked(check);
    }

    @Override
    public boolean validate() {
        return true;
    }
}
