package com.smartgwt.mobile.client.widgets.layout;

import java.util.Stack;

import com.google.gwt.event.logical.shared.ValueChangeEvent;
import com.google.gwt.event.logical.shared.ValueChangeHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.user.client.History;
import com.google.gwt.user.client.ui.Widget;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.core.Function;
import com.smartgwt.mobile.client.internal.util.AnimationUtil;
import com.smartgwt.mobile.client.internal.util.AnimationUtil.Direction;
import com.smartgwt.mobile.client.internal.widgets.events.BeforePanelShownEvent;
import com.smartgwt.mobile.client.internal.widgets.events.BeforePanelShownHandler;
import com.smartgwt.mobile.client.internal.widgets.events.HasBeforePanelShownHandlers;
import com.smartgwt.mobile.client.internal.widgets.layout.NavigationBarItem;
import com.smartgwt.mobile.client.types.PanelHeaderPosition;
import com.smartgwt.mobile.client.util.SC;
import com.smartgwt.mobile.client.widgets.Action;
import com.smartgwt.mobile.client.widgets.ActionContext;
import com.smartgwt.mobile.client.widgets.Canvas;
import com.smartgwt.mobile.client.widgets.ContainerFeatures;
import com.smartgwt.mobile.client.widgets.Header1;
import com.smartgwt.mobile.client.widgets.NavigationButton;
import com.smartgwt.mobile.client.widgets.Panel;
import com.smartgwt.mobile.client.widgets.PanelContainer;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.client.widgets.events.ClickHandler;
import com.smartgwt.mobile.client.widgets.events.HasPanelHideHandlers;
import com.smartgwt.mobile.client.widgets.events.HasPanelShowHandlers;
import com.smartgwt.mobile.client.widgets.events.HasPanelSkipHandlers;
import com.smartgwt.mobile.client.widgets.events.PanelHideEvent;
import com.smartgwt.mobile.client.widgets.events.PanelHideHandler;
import com.smartgwt.mobile.client.widgets.events.PanelShowEvent;
import com.smartgwt.mobile.client.widgets.events.PanelShowHandler;
import com.smartgwt.mobile.client.widgets.events.PanelSkipEvent;
import com.smartgwt.mobile.client.widgets.events.PanelSkipHandler;
import com.smartgwt.mobile.client.widgets.tab.TabSet;

public class NavStack<T extends Canvas & NavigationBarItem> extends Panel implements PanelContainer, HasBeforePanelShownHandlers,
HasPanelShowHandlers, HasPanelHideHandlers, HasPanelSkipHandlers, ValueChangeHandler<String> {

    protected Stack<Panel> stack = new Stack<Panel>();
	protected NavigationBar navigationBar = new NavigationBar(this);
    private ScrollableLayout layout;
    private boolean animating = false;
    private Function<Void> resetAnimatingFlag = null;

    {
        layout  = new ScrollableLayout(44, 0);
        getElement().addClassName("sc-navstack");
        if (_HISTORY_ENABLED) History.addValueChangeHandler(this);
    }

	public NavStack() {
	}

	public NavStack(String title) {
		super(title);
	}

	public NavStack(String title, ImageResource icon) {
		super(title,icon);
	}

	public NavStack(Panel panel) {
		super(panel.getTitle(),panel.getIcon());
		push(panel);
	}

    @Override
    public void onLoad() {
        super.onLoad();
        Widget parentWidget = getParent();
        if (parentWidget != null && parentWidget.getParent() instanceof TabSet) {
            layout.setBottom(0);
        } else {
            layout.setBottom(0);
        }
    }

    @Override
    public void setVisible(boolean visible) {
        super.setVisible(visible);
        if (visible) {
            navigationBar._layOutMembers();
        }
    }

	public NavigationBar getNavigationBar() {
		return navigationBar;
	}

    @SGWTInternal
    public final boolean _isAnimating() {
        return animating;
    }

    private void setAnimatingFlag() {
        resetAnimatingFlag = new Function<Void>() {
            @Override
            public Void execute() {
                if (resetAnimatingFlag == this) animating = false;
                return null;
            }
        };
        animating = true;
    }

    protected void pushHistory() {
        if (!_HISTORY_ENABLED) return;
        String historyToken = History.getToken();
        SC._debug("pushHistory historyToken="+historyToken);
        if(stack.size()>1) {
            String newHistory = buildHistory(stack.size());
            if(!historyToken.equals(newHistory)) {
                if(newHistory.length() > historyToken.length()) {
                    SC._debug("pushHistory adding new item to history currentHistory="+historyToken+" newHistory="+newHistory);
                    History.newItem(newHistory);
                } else {
                    stack.peek().parseHistory(this);
                }
            }
        } else {
            parseHistory();            
        }
    }
    
    protected void parseHistory() {
    }
    
    protected void popHistory() {
        if (!_HISTORY_ENABLED) return;
        String history = buildHistory(stack.size());
        SC._debug("popHistory setting history="+history);
        History.newItem(history);
    }
    
    protected String buildHistory(int depth) {
        String currentHistory = "";
        if(stack.size() >= depth) {
            for(int i = 0; i < depth; ++i) {
                currentHistory += stack.get(i).getTitle();
                if(i < depth-1) {
                    currentHistory += "/";
                }
            }
        }
        return currentHistory;
    }

	public void push(Panel panel) {
	    push(panel, Direction.LEFT);
	}

	protected void push(final Panel panel, Direction direction) {
		final NavStack<T> navStack = this;	
        if(panel.isShowPanelHeader() && direction == Direction.UP) {           
            Panel oldPanel = stack.peek();
            setAnimatingFlag();
            AnimationUtil.slideTransition(oldPanel, panel, layout, direction, new Function<Void>() {
                @Override
                public Void execute() {
                    BeforePanelShownEvent.fire(NavStack.this, panel);
                    return null;
                }
            }, resetAnimatingFlag);
            ((ScrollableLayout)layout).setTop(0);
            PanelShowEvent.fire(NavStack.this, panel);
            PanelHideEvent panelHiddenEvent = new PanelHideEvent(oldPanel);
            fireEvent(panelHiddenEvent);
            panel.containerChanged(this);
            stack.push(panel);
        } else {
    		panel.setHeaderPosition(PanelHeaderPosition.BOTTOM);
    		if(stack.size() > 0) {
    			// animate the new panel into place
    			Panel oldPanel = stack.peek();
                navigationBar.<T>_push(panel);
    			// note: slideTransition automatically adds new member and removes old
                setAnimatingFlag();
                AnimationUtil.slideTransition(oldPanel, panel, layout, direction, new Function<Void>() {
                    @Override
                    public Void execute() {
                        BeforePanelShownEvent.fire(NavStack.this, panel);
                        return null;
                    }
                }, resetAnimatingFlag);
    			// set up a handler to have the back button navigate back
    			navigationBar.getLeftButton().addClickHandler(new ClickHandler() {
    				public void onClick(ClickEvent event) {
                        if (!animating) navStack.pop();
    				}
    			});
                PanelShowEvent.fire(NavStack.this, panel);
    			PanelHideEvent panelHiddenEvent = new PanelHideEvent(oldPanel);
    			fireEvent(panelHiddenEvent);
    		} else {
    			this.addMember(navigationBar);
    			layout.addMember(panel);
    			this.addMember(layout);
    	        navigationBar.setTitleLabel(new Header1(panel.getTitle()));
                BeforePanelShownEvent.fire(this, panel);
                PanelShowEvent.fire(this, panel);
    		}
            panel.containerChanged(this);
            if(panel.getActions() != null) {
                for(int i = 0; i < panel.getActions().length; ++i) {
                    final Action action = panel.getActions()[i];
                    final NavigationButton button = new NavigationButton(action.getTitle());
                    button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            action.execute(new ActionContext() {
                                @Override
                                public Panel getPanel() {
                                    return panel;
                                }
                                @Override
                                public Canvas getControl() {
                                    return button;
                                }                           
                            });
                        }                   
                    });
                    navigationBar.setRightButton(button);                
                }
            }
    		stack.push(panel);
            pushHistory();
        }
	}

    public Panel pop() {
        return pop(Direction.RIGHT);
    }

    public Panel pop(Direction direction) {
        final Panel topPanel = stack.peek();

        final Panel priorPanel = stack.pop();
        final Panel newPanel = stack.peek();
        assert priorPanel == topPanel;

        final Function<Void> beforeCallback = new Function<Void>() {
            @Override
            public Void execute() {
                BeforePanelShownEvent.fire(NavStack.this, newPanel);
                return null;
            }
        };
        final Function<Void> afterCallback = new Function<Void>() {
            @Override
            public Void execute() {
                resetAnimatingFlag.execute();
                PanelShowEvent.fire(NavStack.this, newPanel);
                PanelHideEvent panelHiddenEvent = new PanelHideEvent(priorPanel);
                fireEvent(panelHiddenEvent);
                return null;
            }
        };

        if (topPanel.isShowPanelHeader() && direction == Direction.DOWN) { 
            ((ScrollableLayout)layout).setTop(47);
            setAnimatingFlag();
            AnimationUtil.slideTransition(priorPanel, newPanel, layout, direction, beforeCallback, afterCallback);
        } else {
            popHistory();
            // note: slideTransition automatically adds new member and removes old
            setAnimatingFlag();
            AnimationUtil.slideTransition(priorPanel, newPanel, layout, direction, beforeCallback, afterCallback);
            navigationBar.<T>_pop();
        }
        return topPanel;
    }

    /**
     * Pops off panels until <code>panel</code> is in view.  <code>panel</code> is guaranteed to
     * be visible after this method completes, even if it means that the stack of panels in this
     * <code>NavStack</code> is emptied and then <code>panel</code> is pushed onto the stack.
     * 
     * <p>For each panel that is skipped over (that is, it was not the visible panel, but it also
     * will not end up being the visible panel after this method completes), a
     * {@link com.smartgwt.mobile.client.widgets.events.PanelSkipEvent} is fired.  After
     * all the {@link com.smartgwt.mobile.client.widgets.events.PanelSkipEvent}s are fired,
     * there will be a {@link com.smartgwt.mobile.client.widgets.events.PanelShowEvent} fired
     * for the new panel in view and a {@link com.smartgwt.mobile.client.widgets.events.PanelHideEvent}
     * fired for the previously visible panel.
     * 
     * @param panel the panel that should be visible after the operation completes.
     * @return the previously-visible panel, or <code>null</code> if there was no previously
     * visible panel or the currently visible panel is <code>panel</code>.
     */
    public Panel popTo(Panel panel) {
        return popTo(panel, Direction.RIGHT);
    }

    public Panel popTo(final Panel panel, Direction direction) {
        if (! stack.isEmpty()) {
            final Panel topPanel = stack.peek();
            if (topPanel == panel) return null; // Nothing needs to be done because the navstack is already at `panel`.

            Panel priorPanel = null, newPanel = null;
            do {
                priorPanel = stack.pop();
                if (stack.isEmpty()) {
                    newPanel = null;
                } else {
                    newPanel = stack.peek();
                }

                if (priorPanel != topPanel) {
                    navigationBar.<T>_pop();
                    PanelSkipEvent panelSkippedEvent = new PanelSkipEvent(priorPanel);
                    fireEvent(panelSkippedEvent);
                }
            } while (! stack.isEmpty() && newPanel != panel);

            if (! stack.isEmpty()) {
                // We've found the panel.  Apply the transition from `topPanel` to `newPanel`, which equals `panel`.
                assert newPanel == panel;
                priorPanel = topPanel;
                final Panel finalNewPanel = newPanel, finalPriorPanel = priorPanel;
                // note: slideTransition automatically adds new member and removes old
                setAnimatingFlag();
                AnimationUtil.slideTransition(priorPanel, newPanel, layout, direction, new Function<Void>() {
                    @Override
                    public Void execute() {
                        BeforePanelShownEvent.fire(NavStack.this, finalNewPanel);
                        return null;
                    }
                }, new Function<Void>() {
                    @Override
                    public Void execute() {
                        resetAnimatingFlag.execute();
                        PanelShowEvent.fire(NavStack.this, finalNewPanel);
                        PanelHideEvent panelHiddenEvent = new PanelHideEvent(finalPriorPanel);
                        fireEvent(panelHiddenEvent);
                        return null;
                    }
                });
                navigationBar.<T>_pop();
            } else {
                layout.removeMember(priorPanel);
                if (direction == null) push(panel);
                else push(panel, direction.getOppositeDirection());
                PanelHideEvent panelHiddenEvent = new PanelHideEvent(topPanel);
                fireEvent(panelHiddenEvent);
            }
            return topPanel;
        } else {
            assert stack.isEmpty();
            if (direction == null) push(panel);
            else push(panel, direction.getOppositeDirection());
            return null;
        }
    }

	public Panel get(int i) {
		return stack.get(i);
	}

    public void clear() {
        if (animating) throw new IllegalStateException("The NavStack cannot be cleared while a transition animation is in progress.");
        if (!stack.isEmpty()) {
            removeMember(layout);
            final Panel topPanel = top();
            layout.removeMember(topPanel);
            removeMember(navigationBar);

            PanelHideEvent.fire(this, topPanel);

            while (!stack.isEmpty()) {
                PanelSkipEvent.fire(this, stack.pop());
            }

            navigationBar._clear();
        }
    }

	public Panel top() {
		return stack.peek();
	}

	public int size() {
		return stack.size();
	}

	public void setSinglePanel(Panel panel) {
		clear();
        push(panel);
	}

	@Override
	public ContainerFeatures getContainerFeatures() {
		return new ContainerFeatures(false, true, true, true, 1);
	}

    @Override
    public HandlerRegistration _addBeforePanelShownHandler(BeforePanelShownHandler handler) {
        return addHandler(handler, BeforePanelShownEvent.getType());
    }

	@Override
	public HandlerRegistration addPanelHideHandler(PanelHideHandler handler) {
        return this.addHandler(handler, PanelHideEvent.getType());
	}

	@Override
	public HandlerRegistration addPanelShowHandler(PanelShowHandler handler) {
        return this.addHandler(handler, PanelShowEvent.getType());
	}

    @Override
    public HandlerRegistration addPanelSkipHandler(PanelSkipHandler handler) {
        return addHandler(handler, PanelSkipEvent.getType());
    }

    @Override
    public void onValueChange(ValueChangeEvent<String> event) {
        String value = event.getValue();
        if(stack.size() > 0) {
            int size = stack.size();
            String backOne = buildHistory(size-1);
            if(value != null && value.length() > 0 && value.equals(backOne)) {
                SC._debug("onValueChange value="+value+" backOne="+backOne);
                pop();
            } else {
                SC._debug("onValueChange value="+value); 
                parseHistory();
            }
        } 
    }
}
