/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */

package com.smartgwt.mobile.client.widgets.toolbar;

import com.smartgwt.mobile.client.widgets.BaseButton;

/**
 * This class represents a button that is placed on the {@link ToolStrip}.
 * It can either be a back / forward button, or an action button.
 *
 * @see com.smartgwt.mobile.client.widgets.toolbar.ToolStrip#addButton(ToolStripButton) 
 */
public class ToolStripButton extends BaseButton {

    private ButtonType buttonType = ButtonType.BORDERED;
    private boolean inheritTint;

    public ToolStripButton() {
        setButtonType(buttonType);
    }

    public ToolStripButton(String title) {
        this();
        setTitle(title);
    }

    public ToolStripButton(String title, ButtonType buttonType) {
        this(title);
        setButtonType(buttonType);
    }

    public final ButtonType getButtonType() {
        return buttonType;
    }

    public void setButtonType(ButtonType buttonType) {
        this.buttonType = buttonType;
        setButtonTypeClassName(buttonType.getValue());
    }

    public final boolean isInheritTint() {
        return inheritTint;
    }

    public void setInheritTint(boolean inheritTint) {
        this.inheritTint = inheritTint;
        if(inheritTint) {
            getElement().addClassName(_tintClassName);
        } else {
            getElement().removeClassName(_tintClassName);
        }
    }
}

